.286	;16-Bit 286, Real Mode coding.
.MODEL TINY	;Produce a .COM
.CODE	
.DATA
.STARTUP

;****************************************************************************
;*                      ATARI ASTEROIDS EMULATOR v3.02	            *
;*	      ------------------------------                      *
;*                     Programming (c) Chris Pile 1997.                     *
;****************************************************************************
;
	include	equates.asm	;Include Equates
;
	cld		;Always UP

	mov	dx,offset error2	;Point to error (not enough mem)

;
;Initialize the virtual 6502 segment and drawlist segment addresses
;
	mov	ax,cs	;Current code seg
	cmp	ax,07001h	;Enough space?
	jae	terminate_error	;Not enough memory
	add	ax,1000h	;Next 64k segment
	mov	seg_6502,ax	;Save 6502 segment
	add	ax,1000h	;Next 64k segment
	mov	draw_seg,ax	;Save drawlist segment
	jmp	set_segs_ok	;Segments OK

terminate_error:	push	dx	 ;Save error message bit

	mov	dx,offset abort	 ;Abort message
	mov	ah,9
	int	21h	 ;Print it

	mov	dx,offset generror ;General error message
	mov	ah,9	 ;DOS Print string
	int	21h	 ;Print it
	pop	dx	 ;Error bit
	mov	ah,9
	int	21h	  ;Print it
	mov	dx,offset generror1 ;End of error message
terminate_err:	mov	ah,9
	int	21h	  ;Print it
terminate:	mov	ax,4C00h	  ;Terminate
	int	21h	  ;The end

;
;Load the ROM images
;
set_segs_ok:	mov	dx,offset rom_filename ;Point to name
	mov	di,offset state_rom    ;Point to address
	mov	si,8192	     ;8192 bytes
	call	load_file	     ;Load it
	jnc	loaded_rom_ok	         ;Loaded it OK
	mov	si,offset rom_file_errors  ;Point to error table
	mov	dx,[si+bx]	         ;Get message address
	push	dx
	mov	dx,offset abort	;Abort message
	mov	ah,9
	int	21h	;Print it
	pop	dx	;Actual error
	jmp	terminate_err	;Finish

;
;Set up 640x480x16 colour packed pixel mode, CLS's and sets palette
;
loaded_rom_ok:	mov	ah,0Fh	              ;Get video mode
	int	10h	              ;Get it
	mov byte ptr [dos_exit+1],al ;Save it

	xor	ah,ah	;For set mode
	push	ax	;Save
	mov	al,012h	;Mode 640x480x16, packed pixel
	int	10h	;BIOS set
	mov	ah,0Fh	;Get video mode
	int	10h	;Get it
	cmp	al,12h	;Did it set OK?
	pop	ax
	jz	set_palette	;Yes
	int	10h	;Restore original mode
	mov	dx,offset error3	;Video error
	jmp	terminate_error	;Terminate with error

;
;Routine:  set_palette
;Job:      Sets VGA palette (first 16 = game, second 16 = frontend)
;
set_palette:	call	vblank	;Wait for vblank, stops fuzz
	mov	si,offset palette
	xor	al,al
	mov	dx,colour_number	;VGA colour port
set_pal_loop:	out	dx,al	;Colour number AL
	inc	dx	;VGA RGB port
	xchg	ah,al
	lodsb
	out	dx,al	;R
	lodsb
	out	dx,al	;G
	lodsb
	out	dx,al	;B
	dec	dx	;VGA colour port
	xchg	ah,al
	inc	al
	cmp	al,32	;Done?
	jb	set_pal_loop	;No, so more

;
;Move the ATARI ROMS (game + state machine) to the 6502 segment
;Fills empty space in 6502 segment with 0's
;DS must equal CS as ROMS live here ES left pointing to the 6502 segment
;
	mov	es,seg_6502	  ;Set to 6502 segment
	xor	di,di	  ;Address 0
	mov	ax,di
	mov	cx,2800h	  ;5000h bytes
	rep	stosw	  ;Fill with 0's segment
	mov	si,offset state_rom ;Point to ROMS, DI = 5000h
	mov	cx,1024	  ;Copy STATE ROM org 5000h
	rep	movsw	  ;2k move, Leave SI at game ROM
	mov	cx,2048	  ;4k Empty space
	rep	stosw	  ;Fill with 0's
	mov	cx,3072	  ;6k Game ROM org 6800h
	rep	movsw	  ;Move it

;
;Routine:  set_int9
;Job:      Sets up our custom INT 9 keyboard handler
;Returns:  Handler set.  Takes ABSOLUTE control, NO pass-back in INT chain
;
set_int9:	cli		;INT's OFF
	mov	ax,3509h	;Get INT 9
	int	21h	      ;Get it
	mov	word ptr old_int_9,bx   ;Save OLD INT 9 offset
	mov	word ptr old_int_9+2,es ;Save OLD INT 9 segment
	mov	ax,2509h	      ;Set INT 9
	mov	dx,offset int_9	;Point to our handler (DS:DX)
	int	21h	;Set it
	sti		;INT's ON

	call	cls_640x480	;Obvious!!!
	call	front_end_pal	;Setup temp message palette

;
;Load a saved highscore table, if there is one.  No errors reported if not
;
	mov	dx,offset highs_filename    ;Pointer
	mov	si,59	          ;Required size v3.00
	mov	di,offset high_table        ;Load address
	call	load_file	          ;Load table
	jnc	loaded_highs_ok	          ;Loaded v3.00 table
	mov	si,53	          ;v1.00. 2.0x table size
	call	load_file	          ;Try and load it

;
;Time to check for a soundcard, and initialise it if found and load samples
;
loaded_highs_ok:	call	get_blaster	     ;Get the blaster string
	jnz	blaster_vars_got	     ;Found it, check vars
;
;Ask if defaults D1 I5 A220 are ok?
;
	mov	si,offset blast_message1 ;Message
	mov	bx,0213h
	mov	cx,14
	call	print_string
	mov	cx,12
	mov	si,offset blast_message2
	mov	bx,0505h
	call	print_string
	mov	cx,11
	mov	si,offset blast_message3
	mov	bx,0813h
	call	print_string
	mov	cx,14
	mov	si,offset blast_message4
	mov	bx,0C00h
	call	print_string

get_response:	mov	al,byte ptr [key_pressed+1]
	cmp	al,49		;N
	jnz	is_it_y
	jmp	start_no_sounds	;Was N
is_it_y:	cmp	al,21	;Y
	jnz	get_response
	mov	dx,220h	;Default ADDR
	mov	bx,0105h	;Default D and I
	jmp	dma_ok	;Start with defaults

blaster_vars_got:	cmp	cl,7	;Got all components?
	jz	got_all_compo	      ;Yes
	mov	si,offset compo_missing ;No
	mov	bx,0304h
	jmp	sound_load_error

got_all_compo:	cmp	bl,15	;No IRQ's above 15 allowed
	jbe	irq_ok

	mov	si,offset blast_bad_irq
sound_set_error:	mov	bx,0305h
sound_load_error:	mov	cx,14

	call	print_string
	mov	bx,0511h
	mov	si,offset no_sound_mess
	call	print_string
	mov	bx,0A1Ah
	mov	cx,11
	mov	si,offset any_key_mess
	call	print_string
	xor	al,al
	mov	byte ptr [key_pressed+1],al
get_any_key:	mov	al,byte ptr [key_pressed+1]
	test	al,al
	jz	get_any_key
	test	al,80h
	jnz	get_any_key
get_any_key_de:	mov	al,byte ptr [key_pressed+1]
	test	al,80h
	jz	get_any_key_de
	jmp	start_no_sounds

irq_ok:	cmp	bh,3	;DMA 0-3 allowed
	jbe	dma_ok
	mov	si,offset blast_bad_dma
	jmp	sound_set_error

;
;BLASTER variables OK, save 'em and load in the SAMPLE file
;
dma_ok:	push	dx	;Save ADDR
	push	bx	;Save I and D

	call	cls_640x480	;Obvious!

	mov	dx,offset samples_filename  ;Pointer
	mov	si,42447	          ;Required size
	mov	di,offset beat1	          ;Load address
	call	load_file	          ;Load samples
	jnc	samples_loaded	          ;Loaded OK
	pop	ax
	pop	ax	          ;Junk these
	mov	si,offset sam_file_errors   ;Point to error
display_init_err:	mov	si,[si+bx]	          ;Get error message
	mov	bx,[si]	          ;Get XY
	inc	si
	inc	si
	jmp	sound_load_error	          ;Show error

samples_loaded:	pop	bx	;Restore I and D
	pop	dx	;Restore ADDR

	add	dx,6	;Calc DSP reset
	mov	dsp_reset,dx	;Save
	add	dx,4	;Calc DSP read
	mov	dsp_read,dx	;Save
	inc	dx
	inc	dx	;Calc DSP write
	mov	dsp_write,dx	;Save it
	inc	dx
	inc	dx	;Calc DSP ready
	mov	dsp_ready,dx	;Save it

	mov	ah,bh	;Save DMA
	mov	ch,ah	;and again

	or	bh,58h	;Calc DMA mask
	mov	the_dma_mode,bh	;Save it

	mov	byte ptr [doing_high_ints+1],0  ;Turn OFF

	mov	al,11111110b	;Shifter for IRQ mask
	mov	cl,bl	;Get IRQ

	add	bl,8	;Calc INTERRUPT address (0-7)

	cmp	cl,7	;Doing INTs 0-7?
	jbe	low_ints	;Yes
	sub	cl,8	;Adjust for higher INTs
	sub	bl,16
	add	bl,70h	;Adjust for higher INTs

	mov	byte ptr [doing_high_ints+1],bl  ;Turn ON

low_ints:	rol	al,cl	;Shift mask
	mov	irq_mask,al	;Save it
	mov	dsp_irq,bl	;Save it

	mov	al,ah	;Get original DMA number
	or	ah,4	       ;Create mask
	mov byte ptr [dma_mask_val+1],ah ;Save it

	add	al,al	;Double for address port
	cbw		;Zero AH

	mov word ptr [dma_address+1],ax ;Save it
	inc	al	      ;Up one for length port
	mov word ptr [dma_length+1],ax  ;Save it

	mov	bl,ch	   ;Original val back
	mov	bh,ah	   ;Zero hi
	add	bx,offset dma_0	   ;Point to page value
	mov	al,[bx]	   ;Grab it
	mov word ptr [dma_page+1],ax ;Save it

	call	sound_init	;Initialise soundcard
	test	al,al
	jz	start_no_sounds	;Initialised OK

	dec	al	   ;Dec error
	cbw		   ;Zero AH
	mov	bx,ax	   ;Make bx=error code
	mov	si,offset init_e_tab ;Point to table
	jmp	display_init_err	   ;Show error and continue

start_no_sounds:	mov	have_sound,al
	mov	byte ptr [any_sounds+1],al

;
;Routine:  set_int8
;Job:      Sets up our custom INT 8 timer, used for speed crippling
;Returns:  Handler set.
;
set_int8:	cli		;Ints off for this one
	mov	ax,3508h	;Get INT 8
	int	21h	      ;Get it
	mov	word ptr old_int_8,bx   ;Save OLD INT 8 offset
	mov	word ptr old_int_8+2,es ;Save OLD INT 8 segment
	mov	ax,2508h	      ;Set INT 8
	mov	dx,offset int_8	;Point to our handler (DS:DX)
	int	21h	;Set it
	mov	al,34h	;Timer mode 0
	out	43h,al
	mov	al,0AEh	;Low
	out	40h,al
	mov	al,4Dh	;High, sets timer at 60hz
	out	40h,al
	xor	ax,ax	        ;Zero BIOS timer ticker
	mov word ptr [bios_count+1],ax    ;Set it
	sti		        ;Back on

	mov	ax,cs
	mov	es,ax	;Make ES=CS

;***************
;jmp emulator_go
;***************


;
;Front end, assume DS=CS ES=CS
;
front_end:	call	silence_buffer	;Shut-up any sound!!

	call	cls_640x480	;CLS
	call	front_end_pal	;Setup front end palette
	
	mov	si,offset front_s	;Print front_end
dump_front:	lodsb
	inc	al
	jz	done_frontend
	inc	al
	jnz	not_line
	lodsw
	mov	bx,ax
	lodsw
	mov	bp,ax
	lodsw
	mov	cx,ax
	lodsw
	mov	di,ax
	lodsb
	push	si
	mov	si,bp
	call	line
	pop	si
	jmp	dump_front
not_line:	dec	al
	dec	al
	mov	bl,al
	lodsb
	mov	bh,al
	lodsw
	mov	cx,ax
	call	print_string
	jmp	dump_front

done_frontend:	mov byte ptr [key_pressed+1],al          ;Clear keys
	
	mov	ah,current_dips   ;Get switches
	mov	bp,8	;8 switches
	mov	cx,0A0Eh	;White on Blue
	mov	bx,3224h	;XY
display_dips:	mov	al,67	;Switch off
	shr	ah,1
	sbb	al,0	;Switch it on if bit set
	call	out_char	;Print it
	inc	bl	;Inc x
	dec	bp
	jnz	display_dips	;Do remainder of switches

;
;Now display current defined keys
;
show_controls:	mov	bx,2A2Dh	 ;YX
	mov	cx,11	 ;Colour
	mov	si,offset left_key ;Point to current key scancodes
	lodsb		 ;Get left key
	call	dump_keyname	 ;Output the keyname
	inc	bh	 ;Inc Y
	lodsb		 ;Get right key
	call	dump_keyname	 ;Output the keyname
	inc	bh	 ;Inc Y
	lodsb		 ;Get fire key
	call	dump_keyname	 ;Output the keyname
	inc	bh	 ;Inc Y
	lodsb		 ;Get thrust key
	call	dump_keyname	 ;Output the keyname
	inc	bh
	inc	bh	 ;Y + 2
	lodsb		 ;Get hyperspace key
	call	dump_keyname	 ;Output the keyname

front_end_keys:	mov	al,have_sound
	test	al,al
	mov	si,offset s_none
	jnz	front_end_one
	mov	al,byte ptr [any_sounds+1]
	mov	si,offset s_active
	test	al,al
	jz	front_end_one
	mov	si,offset s_disabled

front_end_one:	mov	bx,[si]
	inc	si
	inc	si
	mov	cx,8
	call	print_string

	mov	si,offset c_disabled
	mov	al,byte ptr [cripple_off+1]
	test	al,al
	jz	crip_is_off
	mov	si,offset c_active
crip_is_off:	mov	bx,[si]
	inc	si
	inc	si
	call	print_string
	
front_end_keys1:	mov	al,byte ptr [key_pressed+1]
	test	al,80h
	jz	front_end_keys1
	and	al,7Fh

	cmp	al,88	;EXIT? (F12)
	jnz	not_dos_exit

;
;Save highscore table before exiting proper
;
	mov	dx,offset highs_filename  ;Pointer
	mov	ah,3Ch	        ;Create file
	xor	cx,cx	        ;Zero attrib mask
	int	21h	        ;Create it
	jc	main_exit	   ;Error, so exit
	mov	bx,ax	   ;Make BX handle
	mov	ah,40h	   ;File write
	mov	cx,59	   ;59 bytes
	mov	dx,offset high_table ;Address
	int	21h	   ;Write table out
	mov	ah,3Eh	   ;Close file
	int	21h	   ;Close it

;
;Routine:  restore_int9
;Job:      Restores original INT 9 keyboard handler
;
main_exit:	cli		;INT's OFF
	lds	dx,old_int_9	;Get old handler in DS:DX
	mov	ax,2509h	;Set INT 9
	int	21h	;Set it

;
;Routine:  restore_int8
;Job:      Restores original INT 8 timer
;
	mov	al,34h	;Timer mode 0
	out	43h,al
	xor	al,al
	out	40h,al
	out	40h,al
	lds	dx,cs:old_int_8	;Get old handler in DS:DX
	mov	ax,2508h	;Set INT 8
	int	21h	;Set it
	sti		;Back ON

	call	sound_clear	;Kill soundblaster
dos_exit:	mov	ax,3
	int	10h	;Restore video mode
	jmp	terminate	;Finish

;

not_dos_exit:	cmp	al,59	 ;Start windowed? (F1)
	jnz	not_windowed_go	 ;No
	mov byte ptr [windowed+1],al
	jmp	emulator_go

not_windowed_go:	cmp	al,60	;Start full-screen? (F2)
	jnz	not_win_go	;No
	mov byte ptr [windowed+1],0
	jmp	emulator_go

not_win_go:	cmp	al,61	;Sound toggle? (F3)
	jnz	not_sound_tog	;No

              	mov	al,have_sound	;Do we have sound?
	test	al,al
	jz	we_have_sound
	jmp	front_end_keys	;No

we_have_sound:	mov	al,byte ptr [any_sounds+1]
	test	al,al
	mov	ah,1
	jz	back_on
	dec	ah
back_on:	mov	al,ah
	mov	byte ptr [any_sounds+1],al
	xor	al,al
	jmp	done_frontend

not_sound_tog:	cmp	al,62	;Cripple toggle (F4)
	jnz	not_crip_tog	;No
	not byte ptr [cripple_off+1]
	xor	al,al
	jmp	done_frontend

not_crip_tog:	cmp	al,87	;Key define (F11)
	jz	go_define	;Redefine keys

	cmp	al,2	;Doing dips?
	jae	might_dip	;Might
	jmp	front_end_keys	;No
might_dip:	cmp	al,10
	jb	doing_dip	;Yes
	jmp	front_end_keys

doing_dip:	dec	al
	dec	al	;Make 0-7
	mov	cl,1	;Mask
	xchg	al,cl	;Swap
	shl	al,cl	;Shift mask
	xor	current_dips,al   ;Toggle switch
	xor	al,al	;Zero last key
	jmp	done_frontend	;Show switches

;
;Key redefine engine
;
go_define:	mov	bx,2A2Dh	;YX for current key name
	xor	ax,ax                            ;Zero
	mov	word ptr left_key,ax             ;Clear left/right
	mov	word ptr thrust_key,ax           ;Clear thrust/fire
	mov	hyperspace_key,al	               ;Clear hyperspace
	mov	byte ptr [key_pressed+1],al      ;Zero last key
	mov	byte ptr [flash_flag+1],89
	mov	byte ptr [redraw_allowed+1],1
	mov	byte ptr flash_count,20
	mov	bp,offset left_key               ;Point to key list
define_loop:	mov	cx,8
flash_flag:	mov	al,89
	call	dump_keyname
	mov	al,byte ptr [redraw_allowed+1]
	test	al,al
	jnz	no_flash
	inc	al
	mov	byte ptr [redraw_allowed+1],al
	dec	flash_count
	jnz	no_flash
	mov	byte ptr flash_count,20
	xor	byte ptr [flash_flag+1],3
no_flash:	mov	al,byte ptr [key_pressed+1]
	cmp	al,88
	ja	flash_flag
	cmp	al,6	 ;Prevent ESC, 1,2,3 and 4
	jb	flash_flag

	mov	cl,5	 ;5 keys to check
	mov	di,offset left_key ;Point to key table
	repnz	scasb	 ;See if key is already in use
	jz	define_loop	 ;Yes, so ignore it

	mov	[bp],al	;Store it
	mov	al,90	;Blank any flash message
	call	dump_keyname
	mov	al,[bp]	;Get scancode back
	inc	bp	;Next key store
	mov	cl,11	;Yellow
	call	dump_keyname	;Print it
	inc	bh	;Inc Y
	cmp	bh,46	;Double line?
	jb	no_double	;No
	inc	bh	;Inc Y
no_double:	cmp	bh,48	;Done all keys?
	jb	define_loop	;No

debounce_define:	mov	al,byte ptr [key_pressed+1]
	test	al,80h
	jz	debounce_define
	xor	al,al

	jmp	done_frontend	;Back to frontend

;

emulator_go:	call	cls_640x480	;CLS
	call	game_pal	;Setup game palette

windowed:	mov	al,1	;1= Windowed game, 0= Full screen
	test	al,al

;Self mod code, puts a "JP disp" in at address WINDOWED_SCALE

	mov	ax,((full_scale-windowed_scale) -2) * 256 + 0EBh
	jz	doing_full_scrn	;Not a windowed game

;
;Routine:  draw_frame
;Job:      draws frame around play area
;
	mov	bx,61
	mov	cx,640-62
	mov	si,43
	mov	di,si
	mov	al,1
	pusha
	call	line
	popa
	mov	cx,bx
	mov	si,480-44
	pusha
	call	line
	popa
	mov	di,si
	mov	cx,640-62
	pusha
	call	line
	popa
	mov	bx,cx
	mov	si,43
	call	line

	mov	ax,0FBD1h	;SAR DX,1

doing_full_scrn:	mov word ptr [windowed_scale],ax

;
;Make sure self-mod code pointers are set up
;
	mov	ax,1
	mov word ptr [beat_bytes+1],ax      ;Zero sound
	mov word ptr [explosion_bytes+1],ax ;Zero sound
	mov word ptr [pfire_bytes+1],ax     ;Zero sound
	mov word ptr [sfire_bytes+1],ax     ;Zero sound
	mov word ptr [sauce_bytes+1],ax     ;Zero sound
	mov word ptr [thrust_bytes+1],ax    ;Zero sound
	mov word ptr [extra_bytes+1],ax     ;Zero sound
	mov byte ptr [last_swap],ah         ;Last bank
	mov byte ptr [vector_state],ah      ;Last VECGO
	mov byte ptr [key_mask+1],ah        ;Clear keys
	mov byte ptr [key_pressed+1],ah     ;Clear keys
	mov byte ptr [pfire_allowed+1],ah   ;Allow fire
	mov byte ptr [sfire_allowed+1],ah   ;Allow fire
	mov byte ptr [high_score_flag+1],ah ;Restore hi-scores
	mov	al,10h	          ;Beat toggle
	mov byte ptr [beat_toggle+1],al     ;Store it

;
;Set up key scancodes
;
	mov	ax,word ptr left_key       ;Get scancodes left/right
	xchg	ah,al	         ;Swap hi/lo
	mov	word ptr [left_right+1],ax ;Set them
	mov	ax,word ptr thrust_key     ;Get scancodes thrust/fr
	mov	word ptr [fire_thrust+1],ax ;Set them
	mov	al,hyperspace_key	          ;Get hyperspace scancode
	mov	byte ptr [hyperspace+2],al  ;Set it

;
;Set up virtual machine
;
	call	reset_drawlists	;Sets ES to drawlist segment
	call	state_reset	;Reset the ATARI state machine!
	call	reset_6502	;Reset 6502, set DS to 6502 seg
	s6502		;Save current 6502 state

;
;The following suite of routines initialise game switches/variables
;
	xor	ax,ax
	mov byte ptr ds:[2007h],al             ;Selftest
	mov byte ptr ds:[vecgo],al             ;VECGO
	mov byte ptr ds:[bank],al              ;Bankswitch
	mov byte ptr ds:[2002h],al             ;VECHALT
;
;Clear sound vars
;
	mov word ptr ds:[3C00h],ax             ;Saucer
	mov word ptr ds:[3C02h],ax             ;Saucer/thrust
	mov word ptr ds:[3C04h],ax             ;fire/extra life
	mov byte ptr ds:[3A00h],al             ;Beat
	mov byte ptr ds:[3600h],al             ;Explosion

	mov	al,cs:current_dips  ;Dipswitches
	mov	cx,0302h
	mov	ah,al
	and	ah,ch
	mov 	ds:[2803h],ah	;Dipswitch 1-2
	shr	al,cl
	mov	ah,al
	and	ah,ch
	mov	ds:[2802h],ah	;Dipswitch 3-4
	shr	al,cl
	mov	ah,al
	and	ah,ch
	mov	ds:[2801h],ah	;Dipswitch 5-6
	shr	al,cl
	mov	ah,al
	and	ah,ch
	mov 	ds:[2800h],ah	;Dipswitch 7-8

;

any_sounds:	mov	al,1
	test	al,al
	jnz	main_loop	;No sounds

wait_for_buffer:	mov	al,byte ptr cs:[buffer_playing+1]
	test	al,al
	jz	wait_for_buffer
wait_for_buffer1:	mov	al,byte ptr cs:[buffer_playing+1]
	test	al,al
	jnz	wait_for_buffer1

main_loop:	mov	al,byte ptr cs:[any_sounds+1]
	test	al,al
	jz	doing_sounds
	jmp	main_no_sounds

;
;The following undocumented mess is the seven-voice realtime sample mixer!!
;This is probably a very poor way of mixing 8-bit mono samples, but it seems
;to work OK!!!
;
doing_sounds:	mov	bp,ds
	mov	ax,cs
	mov	ds,ax
	mov	dx,80FFh	;Bias to silence adjuster
doing_sound_main:	mov	ax,bytes_shoved
	cmp	ax,512	;Next buffer full?
	jb	stuff_sounds	;No, so mix next sound in
	jmp	main_continue	;Yep, wait for soundcard INT

stuff_sounds:	inc	bytes_shoved
	mov	di,dma_pointer
	mov	al,dh
thrust_bytes:	mov	cx,1
	dec	cx
	jz	explosion_bytes
thrust_point:	mov	si,0
	add	al,[si]
	jnc	no_big_thrust
	sub	al,dh
	jc	mixed_thrust
	mov	al,dl
	jmp	mixed_thrust
no_big_thrust:	sub	al,dh
	jnc	mixed_thrust
	xor	al,al
mixed_thrust:	inc	word ptr [thrust_point+1]
	mov	word ptr [thrust_bytes+1],cx
explosion_bytes:	mov	cx,1
	dec	cx
	jz	pfire_bytes
explosion_point:	mov	si,0
	add	al,[si]
	jnc	no_big_exp
	sub	al,dh
	jc	mixed_exp
	mov	al,dl
	jmp	mixed_exp
no_big_exp:	sub	al,dh
	jnc	mixed_exp
	xor	al,al
mixed_exp:	inc	word ptr [explosion_point+1]
	mov	word ptr [explosion_bytes+1],cx
pfire_bytes:	mov	cx,1
	dec	cx
	jz	sfire_bytes
pfire_point:	mov	si,0
	add	al,[si]
	jnc	no_big_pfire
	sub	al,dh
	jc	mixed_pfire
	mov	al,dl
	jmp	mixed_pfire
no_big_pfire:	sub	al,dh
	jnc	mixed_pfire
	xor	al,al
mixed_pfire:	inc	word ptr [pfire_point+1]
	mov	word ptr [pfire_bytes+1],cx
sfire_bytes:	mov	cx,1
	dec	cx
	jz	sauce_bytes
sfire_point:	mov	si,0
	add	al,[si]
	jnc	no_big_sfire
	sub	al,dh
	jc	mixed_sfire
	mov	al,dl
	jmp	mixed_sfire
no_big_sfire:	sub	al,dh
	jnc	mixed_sfire
	xor	al,al
mixed_sfire:	inc	word ptr [sfire_point+1]
	mov	word ptr [sfire_bytes+1],cx
sauce_bytes:	mov	cx,1
	dec	cx
	jz	beat_bytes
sauce_point:	mov	si,0
	add	al,[si]
	jnc	no_big_sauce
	sub	al,dh
	jc	mixed_sauce
	mov	al,dl
	jmp	mixed_sauce
no_big_sauce:	sub	al,dh
	jnc	mixed_sauce
	xor	al,al
mixed_sauce:	inc	word ptr [sauce_point+1]
	mov	word ptr [sauce_bytes+1],cx
beat_bytes:	mov	cx,1
	dec	cx
	jz	extra_bytes
beat_point:	mov	si,0
	add	al,[si]
	jnc	no_big_beat
	sub	al,dh
	jc	mixed_beat
	mov	al,dl
	jmp	mixed_beat
no_big_beat:	sub	al,dh
	jnc	mixed_beat
	xor	al,al
mixed_beat:	inc	word ptr [beat_point+1]
	mov	word ptr [beat_bytes+1],cx
extra_bytes:	mov	cx,1
	dec	cx
	jz	no_extra
extra_point:	mov	si,0
	add	al,[si]
	jnc	no_big_extra
	sub	al,dh
	jc	mixed_extra
	mov	al,dl
	jmp	mixed_extra
no_big_extra:	sub	al,dh
	jnc	mixed_extra
	xor	al,al
mixed_extra:	inc	word ptr [extra_point+1]
	mov	word ptr [extra_bytes+1],cx

no_extra:	mov	[di],al
	inc	di
	mov	dma_pointer,di
main_continue:	mov	ds,bp

;
;Actual main-loop, start here if no sounds
;
main_no_sounds:	r6502		;Restore 6502 state

	cmp	bp,nmicycle	;Need an NMI yet?
	jae	do_nmi	;Yes, so do it.
	jmp	no_nmi_yet	;No
do_nmi:	sub	bp,nmicycle	;Adjust cycle counter
	call	nmi	;Do the NMI
;
;Time to restore our saved highscore table, if we haven't already!!
;
high_score_flag:	mov	al,0	;Restored yet?
	test	al,al
	jnz	high_scores_done	;Yes, so don't do it again!
	not	al	                  ;Toggle flag
	mov	byte ptr cs:[high_score_flag+1],al  ;Save it

	pusha		;Save all
	push	es	;Save
	push	ds	;Save

	mov	ax,ds
	mov	es,ax	;Make ES=6502 segment
	mov	ax,cs
	mov	ds,ax	;Make DS=Our code segment

	cmp	si,7B40h	;Was NMI address this?
	mov	di,29	;Pointer for version 2 ROM
	mov	al,1	;Version 2 flag
	jnz	restore_build2	;Yes, so version 2 ROM
	dec	di	;Adjust pointer for version 1 ROM
	dec	al	;Version 1 flag = 0

restore_build2:	mov	byte ptr [save_version+1],al ;Save version

	mov	si,offset high_table  ;Point to saved scores
	mov	ax,es:[di+20]	    ;Save this word
	mov	dl,es:[di+22]	;and this byte!

	mov	bx,di	;Save pointer
	mov	cx,53	;53 bytes to copy
	rep	movsb	;Copy 'em!

	pop	ds	;Restore
	pop	es	;Restore

	mov	[bx+20],ax	;Restore this word
	mov	[bx+22],dl	;and this byte!

	popa		;Restore all

;
;A few sound engines live here in the NMI caller
;
high_scores_done:
	cmp byte ptr ds:[3600h],3Dh
	jnz	no_large_ex
	mov word ptr cs:[explosion_point+1],offset lexplode
	mov word ptr cs:[explosion_bytes+1],lexplode_length
	dec byte ptr ds:[3600h]
	jmp	no_explo
no_large_ex:	cmp byte ptr ds:[3600h],0FDh
	jnz	no_medium_ex
	mov word ptr cs:[explosion_point+1],offset mexplode
	mov word ptr cs:[explosion_bytes+1],mexplode_length
	dec byte ptr ds:[3600h]
	jmp	no_explo
no_medium_ex:	cmp byte ptr ds:[3600h],0BDh
	jnz	no_explo
	mov word ptr cs:[explosion_point+1],offset sexplode
	mov word ptr cs:[explosion_bytes+1],sexplode_length
	dec byte ptr ds:[3600h]
no_explo:
	xor	al,al
	mov	di,1
	cmp byte ptr ds:[3C04h],al
	jnz	might_pfire
	mov byte ptr cs:[pfire_allowed+1],al
	mov word ptr cs:[pfire_bytes+1],di
	jmp	no_pfire
might_pfire:	cmp word ptr cs:[pfire_bytes+1],di
	jnz	no_pfire
pfire_allowed:	mov	ah,0
	test	ah,ah
	jnz	no_pfire
	not byte ptr cs:[pfire_allowed+1]
	mov word ptr cs:[pfire_point+1],offset pfire
	mov word ptr cs:[pfire_bytes+1],pfire_length
no_pfire:
	cmp byte ptr ds:[3C01h],al
   	jnz	might_sfire
	mov byte ptr cs:[sfire_allowed+1],al
	jmp	no_sfire
might_sfire:	cmp word ptr cs:[sfire_bytes+1],di
	jnz	no_sfire
sfire_allowed:	mov	ah,0
	test	ah,ah
	jnz	no_sfire
	not byte ptr cs:[sfire_allowed+1]
	mov word ptr cs:[sfire_point+1],offset sfire
	mov word ptr cs:[sfire_bytes+1],sfire_length
no_sfire:
	cmp byte ptr ds:[3C05h],al
	jz	no_nmi_yet
	cmp word ptr cs:[extra_bytes+1],di
	jnz	no_nmi_yet
	mov word ptr cs:[extra_point+1],offset extraship
	mov word ptr cs:[extra_bytes+1],extra_length

no_nmi_yet:	push	ds:[2800h]	;Save dips
	push	ds:[2802h]	;Save dips

	call	do_6502	;Do one 6502 instruction
	call	do_6502
	call	do_6502
	call	do_6502
	call	do_6502
	call	do_6502
	call	do_6502
	call	do_6502
	call	do_6502
	call	do_6502

	pop	ds:[2802h]	;Restore dips
	pop	ds:[2800h]	;Restore dips

	s6502		;Save 6502 state

;
;The rest of the sound engines
;
beat_toggle:	mov	al,10h
	cmp	ds:[3A00h],al
	jnz	no_beat
	xor	al,4
	mov byte ptr cs:[beat_toggle+1],al
	cmp	al,14h
	mov	di,offset beat2
	mov	ax,beat2_length
	jz	do_beat2
	mov	di,offset beat1
	mov	ax,beat1_length
do_beat2:	mov word ptr cs:[beat_point+1],di
	mov word ptr cs:[beat_bytes+1],ax
no_beat:
	mov	di,1
	mov	al,80h
	cmp byte ptr ds:[3C00h],al
	jnz	no_saucer
	cmp word ptr cs:[sauce_bytes+1],di
	jnz	no_saucer
	test	ds:[3C02h],al
	jnz	big_sauce
	mov word ptr cs:[sauce_point+1],offset smsaucer
	mov word ptr cs:[sauce_bytes+1],smsaucer_length
	jmp	no_saucer
big_sauce:	mov word ptr cs:[sauce_point+1],offset lgsaucer
	mov word ptr cs:[sauce_bytes+1],lgsaucer_length
no_saucer:
	cmp byte ptr ds:[3C03h],al
	jnz	no_thrust
	cmp word ptr cs:[thrust_bytes+1],di
	jnz	no_thrust
	mov word ptr cs:[thrust_point+1],offset thrust
	mov word ptr cs:[thrust_bytes+1],thrust_length
no_thrust:
	mov	al,ds:[vecgo]	    ;Has 6502 requested draw
	cmp	al,cs:[vector_state]  ;Last vecgo value
	jz	no_redraw	    ;Same, so no update
redraw_allowed:	mov	ah,0
	test	ah,ah
	jnz	redraw_allowed	       ;Used for speed crippling
cripple_off:	mov	ah,0	             ;Zero turns off
	mov byte ptr cs:[redraw_allowed+1],ah

	mov	cs:[vector_state],al     ;Save this request value
update:	call	state_machine	       ;Redraw the vector list

no_redraw:
key_pressed:	mov	al,0	  ;Grab keycode
	cmp	al,1	  ;ESC?
	jz	do_exit	  ;Yes!
	cmp	al,5	  ;Pause?
	jnz	stuff_controls	  ;Nope, so no exit yet

;
;Pause routine.
;
	pusha
	call	silence_buffer
pause_debounce:	mov	al,byte ptr cs:[key_pressed+1]
	test	al,80h
	jz	pause_debounce
	cmp	al,5+80h
	jnz	pause_debounce
	xor	al,al
	mov	byte ptr cs:[key_pressed+1],al
	mov	bx,1C0Ch
	mov	cx,15
	mov	si,offset pause_message
	call	print_string
pause_loop:	mov	al,byte ptr cs:[key_pressed+1]
	test	al,al
	jz	pause_loop
	dec	al
	jnz	pause_off
	popa
	jmp	do_exit
pause_off:	cmp	al,4
	jz	pause_loop
	xor	cx,cx
	mov	si,offset pause_message
	call	print_string
	popa
	jmp	stuff_controls


do_exit:	r6502		  ;Tidy stack

;
;Time to save the highscore table!
;
	mov	ax,cs
	mov	es,ax	;Make ES=CS

save_version:	mov	al,0	;Version 0=1 1=2
	mov	si,29	;Pointer for version 2 ROM
	test	al,al
	jnz	save_build2	;No, so version 2 ROM
	dec	si	;Adjust pointer for version 1 ROM

save_build2:	mov	di,offset high_table  ;Point to saved scores
	mov	cx,53	    ;53 bytes to copy
	rep	movsb	    ;Copy 'em!

	mov	ax,cs
	mov	ds,ax	;Make DS=CS
	mov	es,ax	;Make ES=CS

;****************
;jmp main_exit
;****************

	jmp	front_end	  ;Back to front end.

;
;Stuff game control bits in, from KEYMASK
;
stuff_controls:	mov	ah,byte ptr cs:[key_mask+1]
	mov	cl,80h
	mov	al,ah
	and	al,cl
	mov	ds:[2407h],al
	add	ah,ah
	mov	al,ah
	and	al,cl
	mov	ds:[2406h],al
	add	ah,ah
	mov	al,ah
	and	al,cl
	mov	ds:[2004h],al
	add	ah,ah
	mov	al,ah
	and	al,cl
	mov	ds:[2405h],al
	add	ah,ah
	mov	al,ah
	and	al,cl
	mov	ds:[2003h],al
	add	ah,ah
	mov	al,ah
	and	al,cl
	mov	ds:[2403h],al
	add	ah,ah
	mov	al,ah
	and	al,cl
	mov	ds:[2404h],al
	add	ah,ah
	mov	al,ah
	and	al,cl
	mov	ds:[2400h],al

	jmp	main_loop	;Back to main loop

;
;Routine:  int_9
;Job:      Our keyboard handler
;Args:     None, invoked via INT 9
;Returns:  
;
int_9:	push	ax	;Save
	push	bx	;Save
	push	cx	;Save
	push	dx	;Save

	in	al,60h	                ;Get scancode
	mov	byte ptr cs:[key_pressed+1],al    ;Store it
key_mask:	mov	ah,255	                ;Control bitmask

left_right:	mov	bx,1011h
fire_thrust:	mov	dx,1926h
hyperspace:	mov	cx,3980h

	test	al,al
	jnz	some_key
	mov	ah,al
	jmp	stuff_mask

some_key:	test	al,cl	;Key release?
	jz	process_keypress	;No, so process this key
	and	al,7Fh	;Kill top bit

;
;Key control handler.  Very poor...  Could be improved if I could be bothered
;
	cmp	al,bh
	jz	left_button_r
	cmp	al,bl
	jz	right_button_r
	cmp	al,dh
	jz	fire_button_r
	cmp	al,dl
	jz	thrust_button_r
	cmp	al,ch
	jz	hyper_button_r
	cmp	al,2
	jz	start_p1_r
	cmp	al,3
	jz	start_p2_r
	cmp	al,4
	jz	coin_in_r
	jmp	stuff_mask

left_button_r:	and	ah,01111111b
	jmp	stuff_mask
right_button_r:	and	ah,10111111b
	jmp	stuff_mask
fire_button_r:	and	ah,11011111b
	jmp	stuff_mask
thrust_button_r:	and	ah,11101111b
	jmp	stuff_mask
hyper_button_r:	and	ah,11110111b
	jmp	stuff_mask
start_p1_r:	and	ah,11111011b
	jmp	stuff_mask
start_p2_r:	and	ah,11111101b
	jmp	stuff_mask
coin_in_r:	and	ah,11111110b
	jmp	stuff_mask

process_keypress:	cmp	al,bh
	jz	left_button
	cmp	al,bl
	jz	right_button
	cmp	al,dh
	jz	fire_button
	cmp	al,dl
	jz	thrust_button
	cmp	al,ch
	jz	hyper_button
	cmp	al,2
	jz	start_p1
	cmp	al,3
	jz	start_p2
	cmp	al,4
	jz	coin_in
	jmp	stuff_mask

left_button:	or	ah,10000000b
	jmp	stuff_mask
right_button:	or	ah,01000000b
	jmp	stuff_mask
fire_button:	or	ah,00100000b
	jmp	stuff_mask
thrust_button:	or	ah,00010000b
	jmp	stuff_mask
hyper_button:	or	ah,00001000b
	jmp	stuff_mask
start_p1:	or	ah,00000100b
	jmp	stuff_mask
start_p2:	or	ah,00000010b
	jmp	stuff_mask
coin_in:	or	ah,00000001b

stuff_mask:	mov	byte ptr cs:[key_mask+1],ah   ;Save it

	pop	dx	;Restore
	pop	cx	;Restore
	pop	bx	;Restore

	mov	al,20h
	out	20h,al	;Acknowledge interupt
	pop	ax	;Restore
	iret		;Done

;
;Routine:  int_8
;Job:      Our timer
;Args:     None, invoked via INT 8 at 60hz
;Returns:  Signals 'REDRAW IS ALLOWED'.  Used to force 60hz screen updating
;
int_8:	push	ax	              ;Save
	mov byte ptr cs:[redraw_allowed+1],0    ;Signal redraw OK

bios_count:	mov	ax,0	        ;Bios counter
	add	ax,4DAEh	        ;Add our cycles
	mov word ptr cs:[bios_count+1],ax ;Restore
	jc	do_bios_tick	        ;Yes, so do Bios tick
	mov	al,20h
	out	20h,al	        ;Signal EOI
	pop	ax	        ;Restore
	iret		        ;Done

do_bios_tick:	pop	ax	        ;Restore
	jmp dword ptr cs:old_int_8        ;Do BIOS tick

;=============================================================================

;
;Load and save file routines
;
;Routine:  load_file
;Job:      Loads in a file of given name and size to given address in DS
;Args:     DS:DX ASCIIZ filename, SI=required size, DI=required address
;Returns:  NC=OK C=Error, BX=Error message table offset
;
load_file:	push	dx	;Save name pointer
	push	di	;Save address
	push	si	;Save size
	mov	ax,3DC0h	;Open file (read access)
	xor	cl,cl	;Clear attrib mask
	int	21h	;Open file
	jc	open_error	;Error on open
	mov	bx,ax	;Handle to BX
;
;File is open, now seek to end to check length
;
	mov	ax,4202h	;File seek (to end)
	xor	cx,cx	;Zero hi word
	mov	dx,cx	;and lo word
	int	21h	;Do the seek
	jc	seek_error	;Error on seek (same as size err)
	test	dx,dx	;Hi word should be clear
	jnz	seek_error	;Error
	pop	dx
	push	dx	;Restore/save size again
	cmp	ax,dx	;Size the same?
	jnz	seek_error	;No, so error
;
;Rewind pointer and load the file, as the size is correct!
;
	mov	ax,4200h	;File seek (from start)
	xor	cx,cx	;Zero hi word
	mov	dx,cx	;and lo word
	int	21h	;Do the seek
	jc	seek_error	;Seek error
	mov	ah,3Fh	;File read
	pop	cx	;Size to CX
	pop	dx	;Buffer to DX
	push	cx	;Save size
	int	21h	;Do the read
	pop	cx	;Restore size
	jc	read_error	;Error on read
	cmp	ax,cx	;Same as bytes read?
	jnz	read_error	;No so read error
	mov	ah,3Eh	;Close file
	int	21h	;Close it
	pop	dx	;Restore name pointer
	xor	al,al	;Clear carry as all ok
	ret		;Done
;
;File error handlers
;
open_error:	pop	si	;Restore
	pop	di	;Restore
	pop	dx	;Restore
	xor	bx,bx	;Error message pointer
	stc		;Carry
	ret		;Done
seek_error:	mov	ah,3Eh	;Close file
	int	21h	;Close it
	pop	si	;Restore
	pop	di	;Restore
	pop	dx	;Restore
	mov	bx,2	;Error message pointer
	stc		;Carry
	ret		;Done
read_error:	mov	ah,3Eh	;Close file
	int	21h	;Close it
	pop	dx	;Restore
	mov	bx,4	;Error message pointer
	stc		;Carry
	ret		;Done

;=============================================================================

;
;Routine:  cls_640x480
;Job:      Clears the 640x480 screen and sets up packed pixel mode
;Args:     None
;Returns:  None
;
cls_640x480:	push	es	;Save ES
	mov	dx,sequ_addr	;VGA sequencer base
	mov	ax,0F02h	;Index 2, set write all planes
	out	dx,ax
	mov	dx,packed_pix	;Packed pixel base port
	mov	ax,205h	;Index 5, write bits 0-3 = colour
	out	dx,ax
	mov	ax,3h	;Index 3, set write unmodified
	out	dx,ax
	mov	ax,0FF08h	;Index 8, bitmask set to 11111111b
	out	dx,ax
	mov	ax,0A000h	;Video segment
	mov	es,ax	;ES = video segment
	xor	di,di	;Top left of screen
	mov	ax,di	;Colour zero
	mov	cx,19200	;Screen byte size / 2
	rep	stosw	;Clear screen
	pop	es	;Restore ES
;
;Now set up the VGA for packed pixel line drawing.
;Bitmask already set to 255 by the CLS routine.  The data rotate register at
;index 3 was also set to not modify write data.
;
	mov	ax,0305h	;Mode register set to use write
			;data ANDed with bitmask register
	out	dx,ax
	ret

;
;Routine:  vblank
;Job:      Waits for vertical blank
;Args:     None
;Returns:  Nothing
;
vblank:	mov	dx,3DAh	;VGA port
ras1:	in	al,dx
	test	al,8
	jnz	ras1	;Already in vblank, wait to end
ras2:	in	al,dx
	test	al,8
	jz	ras2 	;Refreshing, wait for vblank
	ret

;
;Now we need to set the first 16 palette registers to equal the first 16 VGA
;palette numbers, for packed pixel plotting. Might as well let the BIOS do it!
;
game_pal:	xor	bx,bx	;ZERO register and PALETTE numbers
set_REGs:	call	vblank
set_regs_loop:	push	bx
	mov	ax,1000h	;BIOS set register
	int	10h	;Set it
	pop	bx
	inc	bl	;Next register
	inc	bh	;Next VGA palette number
	cmp	bl,16	;Done 16 of 'em?
	jb	set_regs_loop	;Not yet!
	ret		;Done
;
;Enter here to force the second set of 16 colours.  Used for the front end.
;
front_end_pal:	mov	bx,1000h
	jmp	set_REGs

;
;Routine:  dump_keyname
;Job:      Dumps the name of the key scancode in AL to BH-Y BL-X colour CX
;Args:     The above!
;Returns:  Nothing, registers preserved.
;
dump_keyname:	pusha		      ;Save all
	mov	bp,bx	      ;Save BX
	cbw		      ;Zero high byte
	mov	bx,offset scan_table-1  ;Base pointer
	xlat		      ;Get address offset
	add	ax,offset k_esc	;Add names offset address
	xchg	ax,si	;Make SI the keyword pointer
	mov	bx,bp	;Restore BX

dump_keyname_l:	lodsb		;Get ASCII byte
	test	al,80h	;Test top bit
	jnz	last_character	;Last char if top bit set
	call	out_char	;Print character
	inc	bl	;Inc X
	jmp	dump_keyname_l	;Loop for next character

last_character:	and	al,7Fh	;Mask top bit
	call	out_char	;Print character
	popa		;Restore all
	ret		;Done

;
;Routine:  out_char
;Job:      Prints 1 character on the graphics screen
;Args:     BH = Y 0-59, BL = X 0-79, CH = backcolour, CL = foreground colour
;          AL = ASCII character + 1
;Returns:  Nothing, preserves all registers
;
out_char:	pusha		;Save all regs
	push	es	;Save ES
	push	ds	;Save DS
	mov	dx,0A000h	;Video segment
	mov	es,dx	;Video seg to ES
	mov	dx,cs
	mov	ds,dx	;Make DS = CS for font data

	mov	si,offset font-8	;Point to font data
	xor	ah,ah
	shl	ax,3	;Char value * 8
	add	si,ax	;Add to start of font data

	mov	ah,bh	;Y
	xor	al,al	;Zero al (AL=Y*256)
	mov	bh,al	;Zero bh (Y)
	mov	di,ax
	shr	di,1	;Y*128
	shl	ax,1	;Y*512

	add	di,ax	;DI = Y*640
	add	di,bx	;DI = Y*640+X (screen address)

	mov	dx,packed_pix	;VGA port

	mov	bl,8	;8 pixel rows
	mov	bp,80	;Screen row add
blat_char:	mov	ah,cl	;Foreground colour
	xor	al,al	;Index 0
	out	dx,ax	;Set colour
	lodsb		;Get font data
	mov	bh,al	;Save it
	xchg	al,es:[di]	;Plot foreground bits
	mov	ah,ch	;Background colour
	xor	al,al	;Index 0
	out	dx,ax	;Set colour
	mov	al,bh	;Restore font data
	not	al	;Compliment it
	xchg	al,es:[di]	;Plot background bits
	add	di,bp	;Next screen row
	dec	bl
	jnz	blat_char	;Next character row

	pop	ds	;Restore
	pop	es	;Restore
	popa		;Restore all regs
	ret

;
;Routine:  print_string
;Job:      Prints a string of text to graphic screen
;Args:     CS:SI = zero terminated text string BH = y, BL = x,
;          CH = backcolour, CL foreground colour
;Returns:  Nothing
;
print_string:	push	bx	;Save  X Y
	push	ds	;Save DS
	push	cs
	pop	ds	;Make DS = CS
	xor	dx,dx	;Zero scrambler
string_loop:	lodsb		;Get next char
	test	dh,dh	;Scrambling?
	jnz	descramble	;Yes
	inc	dh	;Signal got KEY
	mov	dl,al	;Save KEY
	jmp	first_char
descramble:	xor	al,dl	;Descramble
	mov	dl,al	;Save KEY
	test	al,al	;Zero.  End of string?
	jz	string_done	;Yes, so done string
first_char:	call	out_char
	inc	bl	;Increase X
	cmp	bl,79	;Wrapping right edge?
	jbe	string_loop	;No.
	xor	bl,bl	;Zero X
	inc	bh	;Increase Y
	cmp	bh,59	;Wrapping bottom?
	jbe	string_loop	;No.
	xor	bh,bh	;Zero Y
	jmp	string_loop	;More chars
string_done:	pop	ds	;Restore DS
	pop	bx	;Restore X Y
	ret

;
;Routine:  LINE   (640x480x16).
;Job:      Draws a line from X1,Y1 to X2,Y2 in chosen colour
;Args:     bx=X1, si=Y1, cx=X2, di=Y2, al=colour (0-15)
;          DS preserved
;Returns:  Nothing, all registers, including BP, are destroyed
;
vertical:	mov	dx,di	;Y2
	sub	dx,si	;Y2 - Y1 = length
	jnc	down_screen	;Already going down
	neg	bp	;Neg ADD
	neg	dx	;Make length positive
down_screen:	inc	dx	;At least one pixel

	add	si,si	 ;Double for table access
	mov	si,cs:[y_table+si] ;Get Y address

	and	cl,7	;Initial pixel pos
	shr	bx,3	;x1/8
	add	si,bx	;si = (y1*80) + (x1/8)
	mov	ah,128	;Left pixel
	ror	ah,cl	;Roll to correct vertical pixel
blit_vertical:	mov	al,ah	;Restore pixel
	xchg	al,[si]	;Plot it
	add	si,bp	;Screen add
	dec	dx	;DEC pixel loop
	jnz	blit_vertical	;Plot remainder
	pop	ds	;Restore
	ret		;Done

line:	push	ds	;Save data segment
	mov	dx,packed_pix	;Packed pixel base port
	mov	ah,al	;Set colour to AH
	xor	al,al	;Index 0, plane fill data set
	out	dx,ax	;Set colour in AH
	mov	ax,0A000h	;Video seg
	mov	ds,ax	;Make DS=Video seg
	mov	bp,80	;Down screen add

	cmp	bx,cx	;x1>x2?
	jz	vertical	;Same, so a vertical line or dot
	jb	left_to_right	;Already left to right

	xchg	bx,cx	;Make x1=x2, x2=x1
	xchg	di,si	;Make y1=y2, y2=y1

left_to_right:	sub	cx,bx	;cx = deltax
	sub	di,si	;di = deltay
	jz	horizontal	;Same, so a horizontal line
	jns	no_neg_deltay	;Deltay already positive
	neg	di	;Make deltay positive
	neg	bp	;Going up screen

no_neg_deltay:	add	si,si	 ;Double for table access
	mov	si,cs:[y_table+si] ;Get Y address

	mov	ah,bl	;Low x1
	and	ah,7	;Initial pixel pos
	shr	bx,3	;x1/8
	add	si,bx	;si = (y1*80) + (x1/8)

	cmp	di,cx	;Deltay > deltax?
	jb	horiz_vert	;No, so mix of horizont/vertical

	mov	bx,cx	;Deltax
	add	bx,bx	;DELT2 = (Deltax * 2)
	mov	dx,cx	;Deltax
	sub	dx,di	;Deltax - Deltay
	add	dx,dx	;ERROR = 2 * (Deltax - Deltay)
	mov	al,128	;Far left pixel
	mov	cl,ah	;Initial pixel pos
	ror	al,cl	;Roll to calc start pixel
	mov	cx,di	;Line length
	inc	cx	;Make sure at least one pixel
	neg	di	;Neg Deltay
	sub	dx,bx	;Adjust ERROR

plot_vert_horiz:	mov	ah,al	;Save pixel
	xchg	ah,[si]	;Plot pixel, needs read/write!!
	add	si,bp	;Add vertical step
	add	di,bx	;Is D variable < 0?
	jns	horiz_step	;No, so a horizontal step needed
	dec	cx	;DEC pixel count
	jnz	plot_vert_horiz	;Loop for next pixel
	pop	ds	;Restore DS
	ret		;Finished
horiz_step:	add	di,dx	;D = D + ERROR
	ror	al,1	;Shift pixel right
	adc	si,0	;Step along X on carry
	dec	cx	;DEC pixel count
	jnz	plot_vert_horiz	;Loop for next pixel
	pop	ds	;Restore DS
	ret		;Finished

horizontal:	inc	cx	;X2 + 1 (at least one pixel)
	mov	dx,cx	

	add	si,si	 ;Double for table access
	mov	si,cs:[y_table+si] ;Get Y address

	mov	cl,bl
	shr	bx,3	;x1/8
	add	si,bx	;si = (y1*80) + (x1/8)
	and	cl,7	;Initial pixel pos
	mov	ax,8000h	;Left pixel and OR mask
	shr	ah,cl	;Shift initial pix pos
do_left_pix:	or	al,ah	;OR in mask
	shr	ah,1	;Next pix
	jc	done_left	;Ready to step
	dec	dx	;DEC length
	jnz	do_left_pix	;More bits
	xchg	al,[si]	;Plot this mask
	pop	ds	;Restore
	ret		;Done
done_left:	dec	dx	;Adjust for last mask shift
	xchg	al,[si]	;Plot left edge
	not	ah	;Right edge mask = 255
	inc	si	;Next screen address
	mov	cl,dl	;Low pixels left
	shr	dx,3	;CX / 8
	jz	do_right_edge	;No middle x8 bits, do right edge
blit_middle:	mov	al,ah	;Set mask to 255
	xchg	al,[si]	;Plot whole group of 8
	inc	si	;Next address
	dec	dx	;DEC length
	jnz	blit_middle	;More middle bytes
do_right_edge:	and	cl,7	;Any right edge?
	jz	no_right_edge	;No, so all done
	xor	al,al	;Initial mask
	ror	ax,cl	;Shift in write bits
	xchg	al,[si]	;Plot right edge group
no_right_edge:	pop	ds	;Restore
	ret		;Done

horiz_vert:	mov	bx,di	;Deltay
	add	bx,bx	;DELT2 = (2 * Deltay)
	mov	dx,di	;Deltay
	sub	dx,cx	;Deltay - Deltax
	add	dx,dx	;ERROR = 2 * (Deltay - Deltax)
	xor 	di,di	;DELT2
	sub	di,cx	;D = DELT2 - Deltax

	inc	cx	;At least one pixel
	mov	al,cl	;Save CL

	mov	cl,ah	;Initial pixel POS
	mov	ah,128	;Far left pixel
	ror	ah,cl	;Roll to calc start pixel
	mov	cl,al	;Restore CL
	xor	al,al	;Zero start mask
	sub	dx,bx	;Adjust ERROR

plot_horiz_vert:	or	al,ah	;Or in current pixel
	add	di,bx	;D = D + DELT2
	jns	vert_step	;D >= 0?
	ror	ah,1	;Add horizontal step
	jc	next_x_address	;On carry, store pixel and step
	dec	cx	;DEC pixel count
	jnz	plot_horiz_vert	;Loop for next pixel
	xchg	al,[si]	;Store any pixels left
	pop	ds	;Restore DS
	ret		;Finished
next_x_address:	xchg	al,[si]	;Plot pixel, needs read/write!!
	inc	si	;Next X address
	xor	al,al	;Zero new pixels
	dec	cx	;DEC pixel count
	jnz	plot_horiz_vert	;Loop for next pixel
	pop	ds	;Restore DS
	ret		;Finished
vert_step:	xchg	al,[si]	;Plot any pixels pending
	xor	al,al	;Zero new pixels
	add	di,dx	;D = D + ERROR
	ror	ah,1	;Do the xstep
	adc	si,bp	;Screen address Y step + X step
	dec	cx	;DEC pixel count
	jnz	plot_horiz_vert	;Loop for next pixel
	pop	ds	;Restore DS
	ret		;Finished

;
;Routine:  CLIPPER (A version of the Cohen-Sutherland viewport algorithm)
;Job:      Clips a line to the visible viewport
;Args:     bx=X1, si=Y1, cx=X2, di=Y2
;Returns:  NZ=invisible, Z=visible with bx, si, cx, di = coords
;
clipper:	xor	ax,ax	;zero endcodes
	cmp	bx,max_x	;is x1 <= max_x?
	jle	clip_x1_ok	;(signed) yes
	or	al,2	;set x1/y1 endcode to 0010
clip_x1_ok:	cmp	bx,min_x	;is x1 >= min_x?
	jge	clip_x1_ok1	;(signed) yes
	or	al,1	;set x1/y1 endcode to 0001
clip_x1_ok1:	cmp	si,max_y	;is y1 <= max_y?
	jle	clip_y1_ok	;(signed) yes
	or	al,4	;set x1/y1 endcode to 0100
clip_y1_ok:	cmp	si,min_y	;is y1 >= min_y?
	jge	clip_y1_ok1	;(signed) yes
	or	al,8	;set x1/y1 endcode to 1000

clip_y1_ok1:	cmp	cx,max_x	;is x2 <= max_x?
	jle	clip_x2_ok	;(signed) yes
	or	ah,2	;set x2/y2 endcode to 0010
clip_x2_ok:	cmp	cx,min_x	;is x2 >= min_x?
	jge	clip_x2_ok1	;(signed) yes
	or	ah,1	;set x2/y2 endcode to 0001
clip_x2_ok1:	cmp	di,max_y	;is y2 <= max_y?
	jle	clip_y2_ok	;(signed) yes
	or	ah,4	;set x2/y2 endcode to 0100
clip_y2_ok:	cmp	di,min_y	;is y2 >= min_y?
	jge	clip_go	;(signed) yes
	or	ah,8	;set x2/y2 endcode to 1000

clip_go:	test	ah,al	;is line invisible?
	jz	visible	;no, so might be visible
	ret		;yes, line invisible

visible:	cmp	ah,al	;is line visible?
	jnz	needs_clipping	;no, so needs clipping
	ret		;yes, line visible

needs_clipping:	test	al,al	;need to clip x1/y1?
	jz	clip_x2y2	;no, so clip x2/y2

clip_x1y1:	mov	dx,min_y	;top edge value
	test	al,8	;clip top edge?
	jnz	clip_x1y1_vert	;yes, so clip it
	mov	dx,max_y	;bottom edge value
	test	al,4	;clip bottom edge?
	jnz	clip_x1y1_vert	;yes, so clip it
	mov	dx,max_x	;right edge value
	test	al,2	;clip right edge?
	jnz	clip_x1y1_horiz	;yes, so clip it
	mov	dx,min_x	;left edge value
clip_x1y1_horiz:	mov	bp,cx	;x2
	sub	bp,bx	;x2-x1
	mov	ax,dx	;xmax/min
	sub	dx,bx	;xmax/min-x1
	mov	bx,ax	;x1=xmax/min
	mov	ax,di	;y2
	sub	ax,si	;y2-y1
	imul	dx	;(y2-y1) * (xmax/min-x1)...
	idiv	bp	;  div (x2-x1)
	add	si,ax	;y1+clip result
	jmp	clipper
clip_x1y1_vert:	mov	bp,di	;y2
	sub	bp,si	;y2-y1
	mov	ax,dx	;ymax/min
	sub	dx,si	;ymax/min-y1
	mov	si,ax	;y1=ymax/min
	mov	ax,cx	;x2
	sub	ax,bx	;x2-x1
	imul	dx	;(x2-x1) * (ymax/min-y1)...
	idiv	bp	;  div (y2-y1)
	add	bx,ax	;x1+clip result
	jmp	clipper

clip_x2y2:	mov	dx,min_y	;top edge value
	test	ah,8	;clip top edge?
	jnz	clip_x2y2_vert	;yes, so clip it
	mov	dx,max_y	;bottom edge value
	test	ah,4	;clip bottom edge?
	jnz	clip_x2y2_vert	;yes, so clip it
	mov	dx,max_x	;right edge value
	test	ah,2	;clip right edge?
	jnz	clip_x2y2_horiz	;yes, so clip it
	mov	dx,min_x	;left edge value
clip_x2y2_horiz:	mov	bp,cx	;x2
	sub	bp,bx	;x2-x1
	mov	cx,dx	;x2=xmax/min
	mov	ax,di	;y2
	sub	ax,si	;y2-y1
	sub	dx,bx	;xmax/min-x1
	imul	dx	;(y2-y1) * (xmax/min-x1)...
	idiv	bp	;  div (x2-x1)
	mov	di,si	;y2=y1
	add	di,ax	;y2+clip result
	jmp	clipper
clip_x2y2_vert:	mov	bp,di	;y2
	sub	bp,si	;y2-y1
	mov	di,dx	;y2=xmax/min
	mov	ax,cx	;x2
	sub	ax,bx	;x2-x1
	sub	dx,si	;ymax/min-y1
	imul	dx	;(x2-x1) * (ymax/min-y1)...
	idiv	bp	;  div (y2-y1)
	mov	cx,bx	;x2=x1
	add	cx,ax	;x2+clip result
	jmp	clipper

;=============================================================================

;
;The following suite of routines emulate the ATARI b&w vector hardware, or
;the 'State Machine' as it's known.
;
;Expects DS=6502 segment, ES=drawlist segment, SS=CS as BP used as index to
;state machine instruction emulation table & variables.
;
;Routine:  reset_drawlists & Hashtable
;Job:      Clears drawlists and resets pointers
;Args:     None, DS must equal CS.  Sets ES to the DRAW SEGMENT
;Returns:  Nothing
;
reset_drawlists:	mov	es,draw_seg	;Set ES to drawlist segment
	mov	ax,0FFFFh	;Terminator
	mov	bx,drawlist_add	;Add value between lists
	mov	di,erase_list	;Point to erase list
	stosw		;Terminate erase list
	add	di,bx	     ;Point to drawlist 1 (32768)
	mov	[current_list_clr],di  ;Set drawlist pointer
	stosw		     ;Terminate drawlist 1
	add	di,bx	;Point to drawlist 2 (49152)
	stosw		;Terminate drawlist 2
;
;Clear the hashtable
;
clear_hashtable:	xor	di,di	;Hashtable at Zero
	mov	cx,8192	;16384 word entries
	mov	ax,di	;Clear index
	rep	stosw	;Clear it
	ret		;Done
	
;
;Routine:  state_machine
;Job:      Runs state machine from address in PC, constructs drawlist in ES
;          from state machine code in DS:PC, also draws the list, erasing any
;          dead lines
;Args:     None
;Returns:  Nothing
;
state_machine:	push	ds	      ;Save DS
	mov 	bp,offset state_codes   ;Set index reg
	call	clear_hashtable	      ;Clear the hashtable
	call	go_state	      ;Call the state machine

;
;Now blit through old drawlist adding any lines that are not in the current
;drawlist to the erase list.  Next erase any dead lines and draw the current
;drawlist.  Finally toggle current drawlist to old drawlist.
;
	mov	ax,es	      ;Drawlist segment
	mov	ds,ax	      ;Make DS = drawlist segment
	mov	si,cs:current_list_clr  ;SI = current drawlist
	push	si	      ;Save current drawlist
	mov	di,erase_list	;Erase list pointer (also toggle!)
	xor	si,di	;SI = old drawlist

search_main:	mov	ax,[si]	;Get X1
	inc	ax	;Set zero on 0FFFFh (Terminator)
	jz	all_done	;Finished search
	dec	ax
	inc	si
	inc	si
	mov	cx,[si]	;Y1
	inc	si
	inc	si
	mov	dx,[si]	;X2
	inc	si
	inc	si
	mov	bp,[si]	;Y2
	inc	si
	inc	si
	inc	si	;Skip colour byte
;
;Calc HASHCODE, see if line is in the new list, if not; put in the erase list
;
	mov	bx,ax	;Save x1
	xor	ax,cx	;x1 XOR y1
	xchg	ah,al
	xor	ax,dx	;XOR x2
	xchg	ah,al
	xor	ax,bp	;XOR y2
	add	ax,ax	;Word value
	xchg	bx,ax	;Index into HASHTABLE (restore x1)

	push	si	;Save our place in old drawlist
hash_search:	and	bx,16383	;HASH modulo
	mov	si,[bx]	;Get value from table (new list)
	test	si,si	;Is this entry free?
	jnz	line_still_lives	;No, so this line might be same
;
;Slot was free, so line is dead.  Add it to the erase list
;
	pop	si	;Restore our place in old drawlist
	mov	[di],ax	;x1
	inc	di
	inc	di
	mov	[di],cx	;y1
	inc	di
	inc	di
	mov	[di],dx	;x2
	inc	di
	inc	di
	mov	[di],bp	;y2
	inc	di
	inc	di
	jmp	search_main	;Do next line in new list
;
;HASH entry wasn't free, so check if lines are same.  If same do the next line
;in oldlist else goto next slot in HASHTABLE and repeat check until free slot
;
line_still_lives:	inc	bx
	inc	bx	;Next hash entry
	cmp	ax,[si]	;Old x1 = New x1?
	jnz	hash_search	;No
	cmp	cx,[si+2]	;Old y1 = New y1?
	jnz	hash_search	;No
	cmp	dx,[si+4]	;Old x2 = New x2?
	jnz	hash_search	;No
	cmp	bp,[si+6]	;Old y2 = New y2?
	jnz	hash_search	;No
;
;Line is in both lists, so it's still alive!  Move to next line in old list
;
	pop	si	;Restore our place in old drawlist
	jmp	search_main	;Do next line

all_done:	dec	ax	;Terminator 0FFFFh in erase list
	mov	[di],ax
;
;Now erase old lines, and draw fresh ones!  Inline VBLANK for speed
;
	mov	si,erase_list	;Erase list

	mov	dx,3DAh	;VGA port
in_vblank:	in	al,dx
	test	al,8
	jnz	in_vblank	;Already in vblank, wait for end
refreshing:	in	al,dx
	test	al,8
	jz	refreshing	;Refreshing, wait for vblank

erase_lines:	mov	bx,[si]
	inc	bx
	jz	erased_all
	dec	bx	;x1
	inc	si
	inc	si
	mov	ax,[si]	;y1
	inc	si
	inc	si
	mov	cx,[si]	;x2
	inc	si
	inc	si
	mov	di,[si]	;y2
	inc	si
	inc	si
	push	si
	mov	si,ax	;y1
	xor	al,al	;Black lines
	call	line
	pop	si
	mov	bx,[si]
	inc	bx
	jz	erased_all
	dec	bx	;x1
	inc	si
	inc	si
	mov	ax,[si]	;y1
	inc	si
	inc	si
	mov	cx,[si]	;x2
	inc	si
	inc	si
	mov	di,[si]	;y2
	inc	si
	inc	si
	push	si
	mov	si,ax	;y1
	xor	al,al	;Black lines
	call	line
	pop	si
	mov	bx,[si]
	inc	bx
	jz	erased_all
	dec	bx	;x1
	inc	si
	inc	si
	mov	ax,[si]	;y1
	inc	si
	inc	si
	mov	cx,[si]	;x2
	inc	si
	inc	si
	mov	di,[si]	;y2
	inc	si
	inc	si
	push	si
	mov	si,ax	;y1
	xor	al,al	;Black lines
	call	line
	pop	si
	mov	bx,[si]
	inc	bx
	jz	erased_all
	dec	bx	;x1
	inc	si
	inc	si
	mov	ax,[si]	;y1
	inc	si
	inc	si
	mov	cx,[si]	;x2
	inc	si
	inc	si
	mov	di,[si]	;y2
	inc	si
	inc	si
	push	si
	mov	si,ax	;y1
	xor	al,al	;Black lines
	call	line
	pop	si
	jmp	erase_lines

erased_all:	pop	si	      ;Restore current drawlist

	mov	ax,si
	xor	ax,erase_list	      ;Toggle drawlists
	mov	cs:current_list_clr,ax  ;Store it

blit_lines:	mov	bx,[si]
	inc	bx
	jz	lines_all_done
	dec	bx	;x1
	inc	si
	inc	si
	mov	dx,[si]	;y1
	inc	si
	inc	si
	mov	cx,[si]	;x2
	inc	si
	inc	si
	mov	di,[si]	;y2
	inc	si
	inc	si
	mov	al,[si]	;Colour
	inc	si
	push	si
	mov	si,dx	;y1
	call	line
	pop	si
	mov	bx,[si]
	inc	bx
	jz	lines_all_done
	dec	bx	;x1
	inc	si
	inc	si
	mov	dx,[si]	;y1
	inc	si
	inc	si
	mov	cx,[si]	;x2
	inc	si
	inc	si
	mov	di,[si]	;y2
	inc	si
	inc	si
	mov	al,[si]	;Colour
	inc	si
	push	si
	mov	si,dx	;y1
	call	line
	pop	si
	mov	bx,[si]
	inc	bx
	jz	lines_all_done
	dec	bx	;x1
	inc	si
	inc	si
	mov	dx,[si]	;y1
	inc	si
	inc	si
	mov	cx,[si]	;x2
	inc	si
	inc	si
	mov	di,[si]	;y2
	inc	si
	inc	si
	mov	al,[si]	;Colour
	inc	si
	push	si
	mov	si,dx	;y1
	call	line
	pop	si
	mov	bx,[si]
	inc	bx
	jz	lines_all_done
	dec	bx	;x1
	inc	si
	inc	si
	mov	dx,[si]	;y1
	inc	si
	inc	si
	mov	cx,[si]	;x2
	inc	si
	inc	si
	mov	di,[si]	;y2
	inc	si
	inc	si
	mov	al,[si]	;Colour
	inc	si
	push	si
	mov	si,dx	;y1
	call	line
	pop	si
	jmp	blit_lines

lines_all_done:	pop	ds	;Restore DS
	ret		;Done

;
;State machine!
;
go_state:	mov	si,4000h	   ;Set state PC
	mov	di,[bp+current_list] ;Point to active drawlist
state_loop:	lodsw		   ;Next state command

	xor	bx,bx	;Zero pointer
	mov	bl,ah	;Get byte containing opcode
	and	bl,0F0h	;Mask out unwanted bits
	shr	bl,3	;Shift opcode for table access
	add	bp,bx	;Add table offset
	mov	dx,[bp]	;Get instruction address
	sub	bp,bx	;Restore index
	jmp	dx	;Do state machine instruction

;
;General area for state machine variables & jump table, indexed via BP
;
	even
state_codes:	dw	illegal_sm	;0 Illegal, treat as HALT
	dw	rel_draw1	;1 Relative draw scale 1 (/256)
	dw	rel_draw2	;2 Relative draw scale 2 (/128)
	dw	rel_draw3	;3 Relative draw scale 3 (/64)
	dw	rel_draw4	;4 Relative draw scale 4 (/32)
	dw	rel_draw5	;5 Relative draw scale 5 (/16)
	dw	rel_draw6	;6 Relative draw scale 6 (/8)
	dw	rel_draw7	;7 Relative draw scale 7 (/4)
	dw	rel_draw8	;8 Relative draw scale 8 (/2)
	dw	rel_draw9	;9 Relative draw scale 9 (/1)
	dw	position_beam	;A Position beam and load scale
	dw	state_halt	;B Halt state machine
	dw	state_call	;C Call subroutine
	dw	state_return	;D Return from subroutine
	dw	state_jmp	;E Jump to address
	dw	short_draw	;F Relative short draw

beam_x	equ	$-state_codes
	dw	0	;Current beam X position
beam_y	equ	$-state_codes
	dw	0	;Current beam Y position
state_stack	dw	10 dup (0)	;State machine return stack
current_list	equ	$-state_codes
current_list_clr	dw	0	;Pointer to current drawlist
state_scale	equ	$-state_codes
	db	0	;Overall state machine scale
state_sp	equ	$-state_codes
	db	0	;Current state machine SP

;
;Individual state machine instruction handlers
;
get_long_line:	mov	cx,ax	;Low Y length to CL
			;Hi Y length +sign +opcode to CH
	and	ch,3	;Mask unwanted bits
	test	ah,4	;Test sign bit
	jz	y_len_positive	;Positive Y length
	neg	cx	;Negative Y length
y_len_positive:	lodsw		;Get X length & intensity
	mov	bx,ax	;Low X length to BL
			;Hi X length +sign +intensity BH
	and	bh,3	;Mask unwanted bits
	test	ah,4	;Test sign bit
	jz	x_len_positive	;Positive X length
	neg	bx	;Negative X length
x_len_positive:	shr	ah,4	;Intensity to low nibble
	mov	dl,ah	;Intensity to DL
	ret		;Done

rel_draw9:	call	get_long_line	;Get initial line from SM codes
store_relative:	push	dx	;Save intensity
			;Bring in overall scale factor
	mov	ax,cx	  ;Tempory Y to AX
	mov	cl,[bp+state_scale] ;Get scale factor
	test	cl,8	  ;Divide or Multiply?
	jnz	divide_scale	;Do divide scale
	and	cl,7	;Mask unwanted div/mult bit
	shl	bx,cl	;Multiply X
	shl	ax,cl	;Multiply Y
	mov	cx,ax	;Restore Y
	jmp	done_scale	;Done
divide_scale:	and	cl,7	;Mask unwanted div/mult bit
	sar	bx,cl	;Divide X
	adc	bx,0	;Add error
	sar	ax,cl	;Divide Y
	adc	ax,0	;Add error
	mov	cx,ax	;Restore Y
			;DL still holds intensity here

done_scale:	add	cx,[bp+beam_y]	;Add relative Y
	add	bx,[bp+beam_x]	;Add relative X

	push	bp	;Save, as used by clipper
	push	si	;Save, as used by clipper
	push	di	;Save, as used by clipper

	mov	si,[bp+beam_y]	;y1

	mov	[bp+beam_y],cx	;New beam_y
	mov	di,cx	;y2
	mov	ax,[bp+beam_x]	;x1
	mov	[bp+beam_x],bx	;New beam_x

    	cmp	dl,3	;Check for black (invisible) line
	jb	not_visible	;Not visible, so don't store

	mov	cx,bx	;x2
	mov	bx,ax	;x1

	call	clipper	;Clip line to screen
	jnz	not_visible	;Not visible, so don't store

;
;Routine:  windowed_scale
;Job:      scales x1,y1,x2,y2 in BX,SI,CX,DI to 640x400 from 0-1023,0-775(?!)
;Args:     BX,SI,CX,DI as x1,y1,x2,y2
;Returns:  scaled x1,y1,x2,y2 in BX,SI,CX,DI and adjusted for window.
;
windowed_scale:	sar	bx,1	;Divide 2
	sar	si,1	;y1/2
	sar	di,1	;y1/2
	sar	cx,1	;Divide 2

;

	neg	di
	add	di,479-46 ;387+6	;Flip Y2 and centre
	neg	si
	add	si,479-46 ;387+6	;Flip Y1 and centre

	add	bx,64	;Centre X1
	add	cx,64	;Centre X2
	jmp	after_scale

not_visible:	pop	di	;Restore
	pop	si	;Restore
	pop	bp	;Restore
	pop	ax	;Throw away intensity
	jmp	state_loop	;Done, so do next instruction
;
;Routine:  full_scale
;Job:      scales x1,y1,x2,y2 in BX,SI,CX,DI to 640x400 from 0-1023,0-775(?!)
;Args:     BX,SI,CX,DI as x1,y1,x2,y2
;Returns:  scaled x1,y1,x2,y2 in BX,SI,CX,DI
;
full_scale:	xor	ax,ax	;Zero adder
	sar	bx,1	;Divide 2
	adc	ax,bx	;x1/2 + (x1 and 1)
	sar	bx,2	;Divide 8
	add	bx,ax	;x1/8 + (x1/2 + x1 and 1)

	xor	ax,ax
	sar	si,1	;y1/2
	adc	ax,si
	sar	ax,2
	add	si,ax

	sar	ax,3
	sub	si,ax
	sar	ax,1
	adc	si,ax

	xor	ax,ax	;Zero adder
	sar	cx,1	;Divide 2
	adc	ax,cx	;x2/2 + (x2 and 1)
	sar	cx,2	;Divide 8
	add	cx,ax	;x2/8 + (x2/2 + x2 and 1)

	xor	ax,ax
	sar	di,1	;y2/2
	adc	ax,di
	sar	ax,2
	add	di,ax

	sar	ax,3
	sub	di,ax
	sar	ax,1
	adc	di,ax

	neg	di
	add	di,479 ;479-22	;Flip Y2 and centre
	neg	si
	add	si,479 ;479-22	;Flip Y1 and centre

;

after_scale:	mov	dx,di	;Y2
	mov	ax,bx	;X1
	mov	bx,si	;Y1

	pop	di	;Restore
	pop	si	;Restore
	pop	bp	;Restore

	push	ax	;Save X1
	push	bx	;Save Y1

;
;Calculate HASH code add add to hashtable, NO full-table checking!!!
;
	xor	ax,bx	;x1 XOR y1
	xchg	ah,al
	xor	ax,cx	;XOR x2
	xchg	ah,al
	xor	ax,dx	;XOR y2
	add	ax,ax	;Word value
	mov	bx,ax	;Index into HASHTABLE
	dec	bx
	dec	bx
find_hash_slot:	inc	bx
	inc	bx
	and	bx,16383	;HASH modulo for "round robin"
	mov	ax,es:[bx]	;Get value from table
	test	ax,ax	;Empty slot?
	jnz	find_hash_slot	;No, so find next slot

	mov	es:[bx],di	;Save index to line in HASHTABLE

	pop	bx	;Restore Y1
	pop	ax	;Restore X1

	stosw		;X1
	mov	ax,bx	;Y1
	stosw
	mov	ax,cx	;X2
	stosw
	mov	ax,dx	;Y2
	stosw
	pop	ax	;Colour (intensity)
	stosb
	jmp	state_loop	;Done, so do next instruction

;
;This looks bad, but it's faster than using some registers and a common
;routine.  Due to the fact that all registers are in use.
;
rel_draw1:	call	get_long_line	;Get initial line from SM codes
	sar	bx,8	;Divide X length 256
	sar	cx,8	;Divide Y length 256
	jmp	store_relative	;Store it in drawlist
rel_draw2:	call	get_long_line	;Get initial line from SM codes
	sar	bx,7	;Divide X length 128
	sar	cx,7	;Divide Y length 128
	jmp	store_relative	;Store it in drawlist
rel_draw3:	call	get_long_line	;Get initial line from SM codes
	sar	bx,6	;Divide X length 64
	sar	cx,6	;Divide Y length 64
	jmp	store_relative	;Store it in drawlist
rel_draw4:	call	get_long_line	;Get initial line from SM codes
	sar	bx,5	;Divide X length 32
	sar	cx,5	;Divide Y length 32
	jmp	store_relative	;Store it in drawlist
rel_draw5:	call	get_long_line	;Get initial line from SM codes
	sar	bx,4	;Divide X length 16
	sar	cx,4	;Divide Y length 16
	jmp	store_relative	;Store it in drawlist
rel_draw6:	call	get_long_line	;Get initial line from SM codes
	sar	bx,3	;Divide X length 8
	sar	cx,3	;Divide Y length 8
	jmp	store_relative	;Store it in drawlist
rel_draw7:	call	get_long_line	;Get initial line from SM codes
	sar	bx,2	;Divide X length 4
	sar	cx,2	;Divide Y length 4
	jmp	store_relative	;Store it in drawlist
rel_draw8:	call	get_long_line	;Get initial line from SM codes
	sar	bx,1	;Divide X length 2
	sar	cx,1	;Divide Y length 2
	jmp	store_relative	;Store it in drawlist

position_beam:	and	ah,15	;Mask out Opcode
	sub	ax,125	;Sub PC bottom of screen offset
	mov	cx,ax	;Beam Y
	lodsw		;Next state machine word
	mov	bl,ah	;Get scale factor into BL
	shr	bl,4	;Shift to low nibble
	test	bl,8	;Test scale div/mult flag
	jz	scale_mult	;Multiply
	neg	bl	;Negate scale count
	or	bl,8	;Set divide flag
	and	bl,15	;Keep scale and div/mult flag
scale_mult:	mov	[bp+state_scale],bl ;Save scale factor
	and	ah,15	  ;Mask unwanted bits
	mov	bx,ax	;Beam X
	mov	[bp+beam_x],bx	;Store beam X
	mov	[bp+beam_y],cx	;Store beam Y
stack_full:	jmp	state_loop	;Done, so do next instruction

state_call:	xor	bh,bh
	mov	bl,[bp+state_sp]	;Get state SP
	cmp	bl,20	;Stack full?
	jae	stack_full	            ;Yes, so can't do call
	mov	cs:[bx+offset state_stack],si ;Stack return address
	inc	bl
	inc	bl	            ;INC SP
	mov	[bp+state_sp],bl              ;Store SP

state_jmp:	and	ah,15	;Mask out opcode
	add	ax,ax	;Double state machine address
call_start:	add	ax,4000h	;Add in 6502 memory space
	mov	si,ax	;Transfer control
	jmp	state_loop	;Do next instruction

state_return:	xor	bh,bh
	mov	bl,[bp+state_sp]	;Get state machine stack pointer
	test	bl,bl	;At bottom of stack?
	jz	state_halt	;No return address so HALT!
	dec	bl
	dec	bl	;DEC SP
	mov	[bp+state_sp],bl	            ;Store SP
	mov	si,cs:[bx+offset state_stack] ;Load return address
	jmp	state_loop	            ;Next instruction

illegal_sm:
state_halt:	mov	es:[di],0FFFFh	    ;Put terminator in drawlist
	ret

;
;Routine:  state_reset
;Job:      Resets the state machine emulator.
;Args:     None
;Returns:  Nothing
;
state_reset:	mov	bp,offset state_codes ;Set index reg
	xor	ax,ax	    ;Centre X,Y
	mov	[bp+beam_x],ax	    ;Centre beam X position
	mov	[bp+beam_y],ax	    ;Centre beam Y position
	mov	[bp+state_scale],al   ;Zero current scale factor
	mov	[bp+state_sp],al	    ;Zero SP
	ret

short_draw:	mov	ch,al	;Save intensity
	shr	ch,4	;Shift intensity to low nibble
	mov	cl,al	;Get scale bit 1
	shr	cl,2	;Shift to bit 1
	and	cl,2	;Just keep bit 1
	xor	bh,bh
	mov	bl,al	;Get X
	and	bl,3	;Mask unwanted bits
	mov	dl,ah	;Get scale bit 0
	shr	dl,3	;Shift to bit 0
	and	dl,1	;Just keep bit 0
	or	cl,dl	;Or in scale bit 1
	xor	dh,dh
	mov	dl,ah	;Get Y
	and	dl,3	;Mask unwanted bits

	inc	cl	;Run scale at x4, x8, x16

	shl	bx,cl	;Scale X
	shl	dx,cl	;Scale Y

	test	al,4	;Test X sign
	jz	x_positive	;X is positive
	neg	bx
x_positive:	test	ah,4	;Test Y sign
	jz	y_positive	;Y is positive
	neg	dx
y_positive:	xchg	cx,dx	;CX = y DH = Intensity
	mov	dl,dh	;Intensity to DL
	jmp	store_relative	;BX = x CX = y

;=============================================================================

;
;The following suite of routines are the soundcard engines
;
;Sound variables
;
	even
old_int_x	dd	0	;Old int X handler (used by SB)
bytes_shoved	dw	0
dma_pointer	dw	0

dsp_reset	dw	0
dsp_read	dw	0
dsp_write	dw	0
dsp_ready	dw	0

the_dma_mode	db	0	;58h + dma channel
dsp_irq	db	0	;IRQ + 8 (or IRQ - 8 + 70h)
irq_mask	db	0	;Bit position of IRQ number
old_PIC0	db	0	;Save PIC0 contents
old_PIC1	db	0	;Save PIC1 contents

have_sound	db	0

;
;DMA page port value tables
;
dma_0:	db	87h
dma_1:	db	83h
dma_2:	db	81h
dma_3:	db	82h

;
;DMA equates
;
dma_mask	equ	0Ah
dma_mode	equ	0Bh
dma_flipflop	equ	0Ch
PIC0	equ	021h
PIC1	equ	0A1h
EOI0	equ	020h
EOI1	equ	0A0h

;
;Routine:  get_blaster
;Job:      Gets information from the BLASTER= environment variable
;Args:     None
;Returns:  Z=variable not found, NZ=found and DX=A var BH=D var BL=I var
;
get_blaster:	push	ds	  ;Save
	push	es	  ;Save
	mov	ds,cs:[2Ch]	  ;Get pointer to environment
	mov	ax,cs
	mov	es,ax	  ;Make ES=CS
	xor	si,si	  ;Zero environment offset
next_string:	mov	di,offset b_string  ;Point to our BLASTER string
	mov	cx,8	  ;Eight chars to compare
get_blast:	mov	al,[si]	;Get char
	cmp	al,' '	;Space?
	jnz	got_first	;No, so continue
	inc	si	;Skip space
	jmp	get_blast	;Next char
got_first:	test	al,al	;Zero?
	jnz	check_string	;No, so continue
	inc	si	;Skip zero
	mov	al,[si]	;Get char
	test	al,al	;Zero?
	jnz	next_string	;No, so continue with next string
	pop	es	;Restore
	pop	ds	;Restore
	ret		;Return ZERO, variable not found

skip_to_upper:	mov	al,[si]	;Get char
	cmp	al,' '	;Space?
	jnz	to_upper	;No, so continue
skip_inc_toupper:	inc	si	;Skip space
	jmp	skip_to_upper	;Repeat check
to_upper:	cmp	al,'a'
	jb	no_convert	;Below 'a' no uppercase conversion
	cmp	al,'z'
	ja	no_convert	;Above 'z' no uppercase conversion
	and	al,223	;Convert lower to uppercase
no_convert:	test	al,al	;Set zero if end of string
	ret		;Done

check_string:	call	skip_to_upper	;Skip spaces and convert uppercase
	cmp	al,es:[di]	;Same as char in our string?
	jz	chars_match	;Yes, so continue
to_zero:	mov	al,[si]	;Get char
	test	al,al	;Zero?
	jz	next_string	;Yes, so check next string
	inc	si	;Next char
	jmp	to_zero	;Repeat check
chars_match:	inc	si	;Next char
	inc	di	;Next char
	loop	check_string	;Loop for 8 characters
;
;Now get the A, D and I components
;
get_components:	cmp	cl,7	;Got all components?
	jnz	more_components	;Not yet!
	jmp	got_all	;Yes!

more_components:	call	skip_to_upper	;Get component
	jz	end_of_variable	;End of string

	cmp	al,'A'	;Address?
	jnz	not_address	;No
	or	cl,2	;Signal A
	call	skip_inc_toupper	;Skip any spaces
	jz	end_of_variable	;End of string
;
;Convert HEX ASCII pointed to by SI into DX
;
	xor	dx,dx	;Zero total
hex_loop:	mov	al,[si]	;Get digit
	cmp	al,'0'
	jb	get_components	;Below 0 so number ends
	cmp	al,'9'
	jbe	hex_num
	call	to_upper	;Above 9 so check for A-F
	cmp	al,'A'
	jb	get_components	;Below A so number ends
	cmp	al,'F'
	ja	get_components	;Above F so number ends
	sub	al,7	;Sub differene
hex_num:	sub	al,'0'	;Sub ASCII code
	shl	dx,4	;Last val * 16
	add	dl,al
	adc	dh,0	;Add in new value
	inc	si	;Next ASCII digit
	jmp	hex_loop	;Do it
;
;Convert the ASCII decimal number pointed to by SI into AH (only 8 bit)
;
get_decimal:	xor	ah,ah	;Zero total
decimal_loop:	mov	al,[si]	;Get digit
	cmp	al,'0'
	jb	decimal_done	;Below 0 so number ends
	cmp	al,'9'
	ja	decimal_done	;Above 9 so number ends
	sub	al,'0'	;Sub ASCII code
	add	ah,ah	;Last val*2
	mov	ch,ah	;Save it
	add	ah,ah
	add	ah,ah	;Last val*8
	add	ah,ch	;Last val*10
	add	ah,al	;Plus new digit
	inc	si	;Next ASCII digit
	jmp	decimal_loop	;Do it
decimal_done:	ret
end_of_variable:	mov	cl,1	;Signal components missing
got_all:	test	cl,cl	;NZ=Found!
	pop	es	;Restore
	pop	ds	;Restore
	ret

b_string:	db	'BLASTER='

not_address:	cmp	al,'D'	;DMA?
	jnz	not_dma	;No
	or	cl,1	;Signal D
	call	skip_inc_toupper	;Skip any spaces
	jz	end_of_variable	;End of string
	call	get_decimal	;Get the DMA number
	mov	bh,ah	;Save it
	jmp	get_components	;Next component

not_dma:	cmp	al,'I'	;IRQ?
	jz	get_irq	;Yes
	inc	si	;Skip this char
	jmp	get_components
get_irq:	or	cl,4	;Signal I
	call	skip_inc_toupper	;Skip any spaces
	jz	end_of_variable	;End of string
	call	get_decimal	;Get the IRQ number
	mov	bl,ah	;Save it
	jmp	get_components

;
;Routine:  sound_init
;Job:      Initialize the SoundBlaster DSP for sample playing.
;Args:     None, takes values from BLASTER env-string.
;Returns:  AL=0 OK, 1=Wrong DSP, 3=INIT failed
;          If OK the IRQ INT will also be set, and the buffers will be reset
;
sound_init:	pushf		;Save INT status
	cli		;INTs off
	push	ds	;Save
	mov	ax,cs
	mov	ds,ax	;Make DS=CS
	mov	dx,dsp_reset	;Point to SB RESET port
	mov	al,1	;RESET request
	out	dx,al	;Send it
	mov	cx,10	;Short wait
init_wait:	in	al,dx	;Wait
	loop	init_wait	;Another
	mov	ax,cx	;Clear RESET request
	out	dx,al	;Send it
	mov	dx,dsp_ready	;Point to READY port
wait_for_dsp_ok:	in	al,dx	;Read it
	test	al,al	;Ready yet?
	js	dsp_is_ready	;YES!
	loop	wait_for_dsp_ok	;No so try again
dsp_fail:	mov	al,3	;Signal SB failed to init
	pop	ds	;Restore
	popf		;Restore INTS
	ret		;Done
dsp_is_ready:	mov	dx,dsp_read	;Point to READ port
	in	al,dx	;Read byte
	cmp	al,0AAh	;Is it this?
	jnz	dsp_fail	;No, so RESET failed
;
;RESET ok, now check the DSP version, AUTO-INIT only works on DSP >=2.00
;
	mov	ah,0E1h	;Get DSP version
	call	send_dsp	;Send it
	call	read_dsp	;Read MAJOR version number
	mov	ah,al	;Save it
	call	read_dsp	;Junk MINOR version.
	cmp	ah,2	;Version >=2.xx ?
	jae	version_ok	;Yes, so procede
	mov	al,1	;Signal wrong DSP
	pop	ds	;Restore
	popf		;Restore INTS
	ret		;Done
;
;Next setup our IRQ interrupt handler, this is called when SB ends playing	
;
version_ok:	push	es	      ;Save ES
	mov	ah,35h	      ;Get INT
	mov	al,dsp_irq	      ;Get IRQ Number
	push	ax	      ;Save IRQ number
	int	21h	      ;Get it
	mov	word ptr old_int_x,bx   ;Save OLD INT x offset
	mov	word ptr old_int_x+2,es ;Save OLD INT x segment
	mov	dx,offset sound_int    ;Point to our handler (DS:DX)
	mov	ah,25h	    ;Set INT x
	mov	al,dsp_irq	    ;IRQ Number
	int	21h	    ;Set it

	pop	ax	    ;Restore IRQ number
	pop	es	    ;Restore ES

	cmp	al,70h	    ;Doing Higher INT's?
	jb	set_low_PIC	    ;No, so set low PIC

	in	al,PIC1	;Get PIC1 contents
	mov	old_PIC1,al	;Save it
	and	al,irq_mask	;CLEAR it (enables)
	out	PIC1,al	;Set SLAVE
	in	al,PIC0	;Get PIC0 contents
	mov	old_PIC0,al	;Save it
	and	al,11111011b	;Set slave enable
	out	PIC0,al	;Enable high INT's
	jmp	INT_all_done	;Done

set_low_PIC:	in	al,PIC0	;Get PIC0 contents
	mov	old_PIC0,al	;Save it
	and	al,irq_mask	;PIC0 IRQ bit
	out	PIC0,al	;CLEAR it (enables)
	in	al,PIC1	;Get PIC1 contents
	mov	old_PIC1,al	;Save it (easy switch off!!)

;
;Setup the DMA code to point to our 1024 byte mix buffer
;
INT_all_done:	mov	bx,offset state_rom
	mov	dx,bx
	mov	ax,cs	;Get segment
	rol	ax,4	;High nibble to low nibble
	mov	ch,al
	and	ch,0Fh	;Page to CH
	and	al,0F0h	;Mask page number
	add	bx,ax	;Add offset
	adc	ch,0	;Bump-up page number if needed
	neg	bx	;Bytes left to page break
	cmp	bx,1024	;Size of buffer
	ja	enough_space	;Enough space in this page
	neg	bx
	add	bx,1025	;Point in next page
	add	dx,1025
	inc	ch	;Up a page
      	neg	bx
enough_space:	neg	bx	      ;Restore pointer
	mov word ptr [dma_buffer+1],dx  ;Save buffer base
	add	dx,512	      ;Point to second half
	mov word ptr [dma_halfway+1],dx ;Save it
	mov	dma_pointer,dx	      ;Save fill pointer
	xor	al,al
	mov byte ptr [buffer_playing+1],al ;Zero toggle
;
;Now setup the DMA controller
;
dma_mask_val:	mov	al,5	;4 + Channel
	out	dma_mask,al	;Set mask
	xor	al,al
	out	dma_flipflop,al	;Clear flip-flop
	mov	al,the_dma_mode	;58h + Channel
	out	dma_mode,al	;Set AUTO-INIT mode

dma_address:	mov	dx,2	;Point to dma_port

	mov	al,bl	;Buffer Low
	out	dx,al
	mov	al,bh	;Buffer Hi
	out	dx,al

dma_page:	mov	dx,83h	;Point to page port

	mov	al,ch	;Page
	out	dx,al

dma_length:	mov	dx,3	;Point to length port

	mov	al,255	;Low
	out	dx,al
	mov	al,3	;Hi    (1024 byte buffer)
	out	dx,al
	mov	al,byte ptr [dma_mask_val+1]  ;DMA channel
	and	al,3	            ;Mask out bit 3
	out	dma_mask,al

	call	silence_buffer_m	;Make sure playback buffer empty

;
;Next, setup playback frequency, in this case 11025 Khz
;
	mov	ah,0D1h	;Speaker on
	call	send_dsp	;Send it
	mov	ah,40h	;Set time constant
	call	send_dsp	;Send it
	mov	ah,165	;256-(1000000 / 11025khz)
	call	send_dsp	;Send it
;
;Setup our block size for the SB, half actual DMA buffer-1 (511)
;
	mov	ah,48h	;Set buffer size
	call	send_dsp	;Send it
	mov	ah,255	;Low
	call	send_dsp	;Send it
	mov	ah,1	;Hi
	call	send_dsp	;Send it
;
;Put SB into AUTO-INIT mode
;
	mov	ah,1Ch	;AUTO-INIT mode
	call	send_dsp	;Send it

	xor	al,al	;Signal all OK
	pop	ds	;Restore
	popf		;Restore INTS
	ret		;Done

;
;Routine:  silence_buffer
;Job:      Fills sound buffer with silence.
;Args:     None
;Returns:  Buffer filled
;
silence_buffer:	mov	al,byte ptr cs:[any_sounds+1]
	test	al,al
	jnz	was_no_sound1

silence_buffer_m:	pushf		;Save INTS
	cli		;Don't want SB cutting in!!
	push	es	;Save ES
	mov	ax,cs
	mov	es,ax	            ;Make ES=CS
	mov	di,word ptr cs:[dma_buffer+1] ;Pointer to buffer
	mov	cx,512	            ;1024 bytes
	mov	ax,08080h	;Silence
	rep	stosw	;Fill buffer
	pop	es	;Restore
	popf		;Restore INTS
was_no_sound1:	ret		;Done

;
;Routine:  send_dsp
;Job:      Sends a byte to the DSP
;Args:     AH byte to send
;Returns:  Nothing
;
send_dsp:	mov	dx,cs:dsp_write	;Point to WRITE port
dsp_w_clear:	in	al,dx	;Read
	test	al,al	;RESET?
	js	dsp_w_clear	;No, so wait for it!
	mov	al,ah	;Byte to send
	out	dx,al	;Send it
	ret		;Done

;
;Routine:  read_dsp
;Job:      Reads a byte from the DSP
;Args:     Nothing
;Returns:  AL byte read
;
read_dsp:	mov	dx,cs:dsp_ready	;Point to READY port
wait_for_ok:	in	al,dx	;Read it
	test	al,al	;Ready yet?
	jns	wait_for_ok	;No!
	mov	dx,cs:dsp_read	;Point to READ port
	in	al,dx	;Grab byte
	ret		;Done

;
;Routine:  sound_clear
;Job:      Kills the soundblaster and it's interrupt
;Args:     None
;Returns:  Nothing
;
sound_clear:	cli		;INT's OFF
	mov	al,byte ptr cs:[any_sounds+1]
	test	al,al
	jnz	was_no_sound

	mov	al,cs:old_PIC0	;Get original PIC0 mask
	out	PIC0,al	;Restore it
	mov	al,cs:old_PIC1	;Get original PIC1 mask
	out	PIC1,al	;Restore it

	mov	ah,0D3h	;Speaker off
	call	send_dsp
	mov	ah,0D0h	;Halt DMA
	call	send_dsp
	mov	ah,0DAh	;Exit AUTO-INIT
	call	send_dsp
	mov	ah,0D0h	;Halt DMA
	call	send_dsp
sound_clear_rst:	mov	dx,cs:dsp_reset	;Point to SB RESET port
	mov	al,1	;RESET request
	out	dx,al	;Send it
	mov	cx,10	;Short wait
clear_rst_wait:	in	al,dx	;Wait
	loop	clear_rst_wait	;Another
	mov	ax,cx	;Clear RESET request
	out	dx,al	;Send it
	lds	dx,cs:old_int_x	;Get old handler in DS:DX
	mov	ah,25h	;Set INT x
	mov	al,cs:dsp_irq	;IRQ number
	int	21h	;Set it
was_no_sound:	sti		;INT's ON
	ret		;Done

;
;Routine:  sound_int
;Job:      Interrupt routine triggered at the end of sample playing
;Args:     None
;Returns:  DMA fill buffer toggled
;
sound_int:	push	ax	;Save
	push	dx	;Save

	xor	ax,ax
	mov	cs:bytes_shoved,ax ;Clear bytes shoved

buffer_playing:	mov	al,0	  ;Toggles on INT
	test	al,al
dma_buffer:	mov	dx,offset state_rom     ;Pointer to buffer start
	jz	fill_first	      ;Fill first half
dma_halfway:	mov	dx,offset state_rom+512 ;Pointer to second half

fill_first:	not	byte ptr cs:[buffer_playing+1]  ;Toggle buffer

	mov	cs:dma_pointer,dx	;Save pointer

	mov	dx,cs:dsp_ready	;Point to DSP_READY
	in	al,dx	;ACK the SB

doing_high_ints:	mov	al,0	;Flag to see if high INT's in use
	test	al,al
	mov	al,20h	;ACK value
	jz	only_low_ack	;No, so no ACK on the SLAVE PIC

	out	EOI1,al	;Signal EOI on SLAVE PIC

only_low_ack:	out	EOI0,al	;Signal EOI on MASTER PIC
 	pop	dx	;Restore
	pop	ax	;Restore
	iret		;Done

;=============================================================================

;
;The following suite of routines emulate a 6502 processor in 32k address space
;
;Routine:  reset_6502
;Job:      Simulates a RESET on the 6502, sets up registers and flags
;Args:     None
;Returns:  8086 Registers set as follows:  DH=6502 flags, DL=6502 Accumulator
;		                 SI=6502 PC,    BX=6502 SP
;		                 CH=6502 X,     CL=6502 Y
;		                 BP=Cycle counter
;DI is used as pointer for 6502 memory accesses
;
reset_6502:	mov	ds,cs:seg_6502	;Set 6502 segment
	mov	si,ds:[07FFCh]	;Set PC to reset vector
	xor	cx,cx	;Zero X, Y registers
	mov	bx,511	;SP to page 1 position FF
	mov	dx,2400h	;INT's OFF and ACCUM to zero
	mov	bp,cx	     ;Zero cycle counter
	mov	byte ptr cs:[nmi+1],dl ;Signal not in NMI
	ret		     ;Reset done

;
;Routine:  do_6502
;Job:      Executes ONE 6502 instruction
;Args:     None, expects DS to be the 6502 segment and the 8086 registers as
;          detailed in RESET above
;Returns:  Nothing
;
do_6502:	mov	al,[si]	     ;Get 6502 instruction
	inc	si
	xor	ah,ah	     ;Zero high
	add	ax,ax	     ;Double for word table access
	add	ax,offset opcode_6502  ;Add opcode table base
	mov	di,ax	     ;DI = table position
	jmp	cs:[di]	     ;Do the 6502 instruction

;
;Individual 6502 instruction handlers
;

irq_test:	test	dh,00000100b	;INT's on?
	jnz	no_nmi_allowed	;No, so no IRQ allowed
	and	dh,11101111b	;Clear BRK flag
irq:	mov	ax,si	;Get PC
	mov	si,07FFEh	;IRQ/BRK vector
go_nmi:	dec	bl
	mov	[bx],ax	;Stack return address
	dec	bl	;Decrease SP
	mov	[bx],dh	;Stack FLAGS
	dec	bl	;Decrease SP
	or	dh,00000100b	;INT's OFF
	mov	si,[si] 	;PC = IRQ vector
	add	bp,7	;7 Cycles
no_nmi_allowed:	ret		;Done

nmi:	mov	al,0	;Alreay in NMI?
	test	al,al
	jnz	no_nmi_allowed	  ;Yes, so don't do NMI
	not	byte ptr cs:[nmi+1] ;Flag in NMI
	mov	ax,si	  ;Get PC
	mov	si,07FFAh	  ;NMI vector
	and	dh,11101111b	  ;Clear BRK flag
	jmp	go_nmi

;
op_00:	;BRK
;
	or	dh,00010000b	;SET BRK flag
	inc	si	;INC PC
	jmp	irq	;Do the BRK

;
op_01:	;ORA (ind,X)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,6	;6 Cycles
	ind_x		;Calc address
	or	dl,[di]	;Do the OR
	jnz	op_01nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_01nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_05:	;ORA (zpage)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,3	;3 Cycles
	zpage		;Calc address
	or	dl,[di]	;Do the OR
	jnz	op_05nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_05nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_06:	;ASL (zpage)
;
	and	dh,01111100b	;Clear sign, zero & carry
	add	bp,5	;5 Cycles
	zpage		;Calculate address
	shl	byte ptr [di],1	;Do the ASL
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_08:	;PHP
;
	add	bp,3	;3 Cycles
	mov	[bx],dh	;Stack flags
	dec	bl	;Decrease SP
	ret		;Done

;
op_09:	;ORA immed
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	mov	al,[si]	;Get immediate value
	inc	si
	or	dl,al	;Do the OR
	jnz	op_09nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_09nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_0A:	;ASL A
;
	inc	bp	
	inc	bp	;2 Cycles
	add	dl,dl	;ASL accum
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_0D:	;ORA (absolute)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	or	dl,[di]	;Do the OR
	jnz	op_0Dnz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_0Dnz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_0E:	;ASL (absolute)
;
	add	bp,6	;6 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	shl	byte ptr [di],1	;Do the ASL
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_10:	;BPL disp
;
	inc	bp
	inc	bp	;2 Cycles
	test	dh,128	;Zero if POSITIVE
	branch_clear	;Return through branch handler

;
op_11:	;ORA (ind),Y
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,5	;5 Cycles
	ind_y		;Calc address
	or	dl,[di]	;Do the OR
	jnz	op_11nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_11nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_15:	;ORA zpage,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	zpage_x		;Calculate address
	or	dl,[di]	;Do the OR
	jnz	op_15nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_15nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_16:	;ASL zpage,X
;
	add	bp,6	;6 Cycles
	zpage_x		;Calculate address
	shl	byte ptr [di],1	;Do the ASL
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_18:	;CLC
;
	inc	bp
	inc	bp	;2 Cycles
	and	dh,11111110b	;Clear CARRY
	ret		;Done

;
op_19:	;ORA abs,Y
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	abs_y		;Calc address
	or	dl,[di]	;Do the OR
	jnz	op_19nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_19nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_1D:	;ORA abs,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	abs_x		;Calc address
	or	dl,[di]	;Do the OR
	jnz	op_1Dnz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_1Dnz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_1E:	;ASL abs,X
;
	add	bp,7	;7 Cycles
	abs_x		;Calculate address
	shl	byte ptr [di],1	;Do the ASL
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_20:	;JSR address
;
	add	bp,6	;6 Cycles
	mov	ax,si	;Get PC
	inc	ax	;Add 1
	dec	bl	;Decrease SP
	mov	[bx],ax	;Stack return address
	dec	bl	;Decrease SP
	mov	si,[si]	;Get call address
	ret		;Done

;
op_21:	;AND (ind,X)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,6	;6 Cycles
	ind_x		;Calculate address
	and	dl,[di]	;Do AND
	jnz	op_21nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_21nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_24:	;BIT (zpage)
;
	add	bp,3	;3 Cycles
	zpage		;Calculate address
	jmp	bit_continue	;Continue as standard BIT A,(nnnn)

;
op_25:	;AND (zpage)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,3	;3 Cycles
	zpage		;Calculate address
	and	dl,[di]	;Do AND
	jnz	op_25nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_25nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_26:	;ROL (zpage)
;
	add	bp,5	;5 Cycles
	zpage		;Calc address
	jmp	rol_continue	;Continue as ROL (absolute)

;
op_28:	;PLP
;
	add	bp,4	;4 Cycles
	inc	bl	;Increase SP
	mov	dh,[bx]	;POP flags
	ret		;Done

;
op_29:	;AND immed
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	mov	al,[si]	;Get immediate value
	inc	si
	and	dl,al	;Do AND
	jnz	op_29nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_29nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_2A:	;ROL A
;
	inc	bp
	inc	bp	;2 Cycles
	mov	al,dh	;AL = 6502 flags
	shr	al,1	;Move 6502's CARRY to 8086's
	adc	dl,dl	;ROLL 6502 accum
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_2C:	;BIT (absolute)
;
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
bit_continue:	mov	al,[di]	;Get contents
	mov	ah,al	;Contents to AH
	and	ah,11000000b	;Keep top two bits
	and	dh,00111101b	;Mask top two bits of flags + Z
	or	dh,ah	;OR top two bits into flags
	and	al,dl	;Do the BIT
	jz	bits_zero	;Set Z flags on ZERO
	ret		;Done
bits_zero:	or	dh,00000010b	;Set ZERO flag
	ret		;Done

;
op_2D:	;AND (absolute)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	and	dl,[di]	;Do AND
	jnz	op_2Dnz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_2Dnz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_2E:	;ROL (absolute)
;
	add	bp,6	;6 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
rol_continue:	mov	al,[di]	;Get byte
	mov	ah,dh	;Get 6502 flags
	shr	ah,1	;Move 6502's CARRY to 8086's
	adc	al,al	;Do the ROL
	mov	[di],al	;Restore byte
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_30:	;BMI disp
;
	inc	bp
	inc	bp	;2 Cycles
	test	dh,128	;Non zero if MINUS flag set
	branch_set	;Return through branch handler

;
op_31:	;AND (ind),Y
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,5	;5 Cycles
	ind_y		;Calculate address
	and	dl,[di]	;Do AND
	jnz	op_31nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_31nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_35:	;AND zpage,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	zpage_x		;Calculate address
	and	dl,[di]	;Do AND
	jnz	op_35nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_35nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_36:	;ROL zpage,X
;
	add	bp,6	;6 Cycles
	zpage_x		;Calc address
	jmp	rol_continue	;Continue as ROL (absolute)

;
op_38:	;SEC
;
	inc	bp
	inc	bp	;2 Cycles
	or	dh,1	;SET carry flag
	ret

;
op_39:	;AND abs,Y
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	abs_y		;Calculate address
	and	dl,[di]	;Do AND
	jnz	op_39nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_39nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_3D:	;AND abs,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	abs_x		;Calculate address
	and	dl,[di]	;Do AND
	jnz	op_3Dnz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_3Dnz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_3E:	;ROL abs,X
;
	add	bp,7	;7 Cycles
	abs_x		;Calc address
	jmp	rol_continue	;Continue as ROL (absolute)

;
op_40:	;RTI
;
	add	bp,6	;6 Cycles
	inc	bl	;Increase SP
	mov	dh,[bx]	;Restore flags
	inc	bl	;Increase SP
	mov	si,[bx]	;Get return address
	inc	bl	;Increase SP
	not	byte ptr cs:[nmi+1] ;Flag NMI finished
	ret		  ;Done

;
op_41:	;EOR (ind,X)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,6	;6 Cycles
	ind_x		;Calc address
	xor	dl,[di]	;Do EOR
	jnz	op_41nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_41nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_45:	;EOR (zpage)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,3	;3 Cycles
	zpage		;Calculate address
	xor	dl,[di]	;Do the EOR
	jnz	op_45nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_45nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_46:	;LSR (zpage)
;
	add	bp,5	;5 Cycles
	zpage		;Calc address
	shr	byte ptr [di],1	;Do the LSR
	flag_sign_zero_c	;Update the S,Z,C flags

;
op_48:	;PHA
;
	add	bp,3	;3 Cycles
	mov	[bx],dl	;Stack ACCUM
	dec	bl	;Decrease SP
	ret		;Done

;
op_49:	;EOR immed
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	xor	dl,[si]	;EOR immediate value
	jnz	op_49nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_49nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	inc	si
	ret		;Done

;
op_4A:	;LSR A
;
	inc	bp
	inc	bp	;2 Cycles
	shr	dl,1	;Do the LSR
	flag_sign_zero_c	;Update the S,Z,C flags

;
op_4C:	;JMP absolute
;
	add	bp,3	;3 Cycles
	mov	si,[si]	;Get JMP address
	ret

;
op_4D:	;EOR (absolute)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	xor	dl,[di]	;Do the EOR
	jnz	op_4Dnz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_4Dnz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_4E:	;LSR (absolute)
;
	add	bp,6	;6 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	shr	byte ptr [di],1	;Do the LSR
	flag_sign_zero_c	;Update the S,Z,C flags

;
op_50:	;BVC disp
;
	inc	bp
	inc	bp	;2 Cycles
	test	dh,64	;Set Zero if V flag clear
	branch_clear	;Return through branch routine

;
op_51:	;EOR (ind),Y
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,5	;5 Cycles
	ind_y		;Calc address
	xor	dl,[di]	;Do EOR
	jnz	op_51nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_51nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_55:	;EOR zpage,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	zpage_x		;Calc address
	xor	dl,[di]	;Do EOR
	jnz	op_55nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_55nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_56:	;LSR zpage,X
;
	add	bp,6	;6 Cycles
	zpage_x		;Calc address
	shr	byte ptr [di],1	;Do the LSR
	flag_sign_zero_c	;Update the S,Z,C flags

;
op_58:	;CLI
;
	inc	bp
	inc	bp	;2 Cycles
	and	dh,11111011b	;Clear INT flag
	ret

;
op_59:	;EOR abs,Y
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	abs_y		;Calc address
	xor	dl,[di]	;Do EOR
	jnz	op_59nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_59nz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_5D:	;EOR abs,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	abs_x		;Calc address
	xor	dl,[di]	;Do EOR
	jnz	op_5Dnz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_5Dnz:	mov	al,dl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_5E:	;LSR abs,X
;
	add	bp,7	;7 Cycles
	abs_x		;Calc address
	shr	byte ptr [di],1	;Do the LSR
	flag_sign_zero_c	;Update the S,Z,C flags

;
op_60:	;RTS
;
	add	bp,6	;6 Cycles
	inc	bl	;Increase SP
	mov	si,[bx]	;Get return address
	inc	bl	;Increase SP
	inc	si	;Point to next instruction
	ret		;Done

;
op_61:	;ADC (ind,X)
;
	add	bp,6	;6 Cycles
	ind_x		;Calculate address
	jmp	adc_continue	;Continue as standard ADC A,(nnnn)

;
op_65:	;ADC (zpage)
;
	add	bp,3	;3 Cycles
	zpage		;Calculate address
	jmp	adc_continue	;Continue as standard ADC A,(nnnn)

;
op_66:	;ROR (zpage)
;
	add	bp,5	;5 Cycles
	zpage		;Calculate address
	jmp	ror_continue	;Continue as standard ROR (nnnn)

;
op_68:	;PLA
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	inc	bl	;Increase SP
	mov	al,[bx]	;POP Accum
	mov	dl,al	;6502 Accum
	test	al,al	;Set flags
	jnz	op_68nz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_68nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_69:	;ADC immed
;
	and	dh,10111111b	;Clear overflow
	inc	bp
	inc	bp	;2 Cycles
	mov	ah,dh	;Set AH = 6502 flags
	test	ah,8	;Doing BCD?
	jnz	op_69bcd	;Yes
	mov	al,[si]	;Get immediate value
	inc	si
	shr	ah,1	;Move 6502 Carry to 8086's
	adc	dl,al	;Do the ADC
	flag_s_z_c_o      ;Done, return setting 6502 flags
op_69bcd:	mov	al,[si]	;Get immediate value
	inc	si
	shr	ah,1	;Move 6502 Carry to 8086's
	adc	dl,al	;Do the ADC
	mov	al,dl
	daa		;Decimal adjust
	mov	dl,al	;8086 accum to 6502's
	flag_s_z_c_o	;Done, return setting 6502 flags

;
op_6A:	;ROR A
;
	inc	bp
	inc	bp	;2 Cycles
	mov	al,dh	;6502 flags to AL
	and	dh,01111100b	;Clear 6502 sign, zero & carry
	shr	al,1	;Move 6502 Carry to 8086's
	rcr	dl,1	;Do the ROR
	adc	dh,0	;Set 6502 carry
	mov	al,dl
	test	al,al	;Set/Res S,Z flags
	jnz	op_6Anz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_6Anz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_6C:	;JMP (address)
;
	add	bp,5	;5 Cycles
	mov	si,[si]	;Get the pointer
	mov	si,[si]	;Do the JMP
	ret

;
op_6D:	;ADC (absolute)
;
	and	dh,10111111b	;Clear overflow
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
adc_continue:	mov	ah,dh	;Set AH = 6502 flags
	test	ah,8	;Doing BCD?
	jnz	op_6Dbcd	;Yes
	shr	ah,1	;Move 6502 Carry to 8086's
	adc	dl,[di]	;Do the ADC
	flag_s_z_c_o	;Done, return setting 6502 flags
op_6Dbcd:	shr	ah,1	;Move 6502 Carry to 8086's
	adc	dl,[di]	;Do the ADC
	mov	al,dl	;8086 accum to 6502's
	daa		;Decimal adjust
	mov	dl,al	;Back to 6502 accum
	flag_s_z_c_o	;Done, return setting 6502 flags

;
op_6E:	;ROR (absolute)
;
	add	bp,6	;6 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
ror_continue:	mov	ah,dh	;Get 6502 flags
	and	dh,01111100b	;Clear 6502 sign, zero & carry
	shr	ah,1	;Move 6502's CARRY to 8086's
	mov	al,[di]	;Get byte
	rcr	al,1	;Do the ROR
	mov	[di],al	;Restore byte
	adc	dh,0	;Set 6502 carry
	test	al,al	;Set/Res S,Z flags
	jnz	op_6Enz	;No ZERO flag set
	or	dh,00000010b	;Set ZERO flag
op_6Enz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_70:	;BVS disp
;
	inc	bp
	inc	bp	;2 Cycles
	test	dh,64	;Set Non Zero if V flag SET
	branch_set	;Return through branch routine

;
op_71:	;ADC (ind),Y
;
	add	bp,5	;5 Cycles
	ind_y		;Calculate address
	jmp	adc_continue	;Continue as standard ADC A,(nnnn)

;
op_75:	;ADC zpage,X
;
	add	bp,4	;4 Cycles
	zpage_x		;Calculate address
	jmp	adc_continue	;Continue as standard ADC A,(nnnn)

;
op_76:	;ROR zpage,X
;
	add	bp,6	;6 Cycles
	zpage_x		;Calculate address
	jmp	ror_continue	;Continue as standard ROR (nnnn)

;
op_78:	;SEI
;
	inc	bp
	inc	bp	;2 Cycles
	or	dh,4	;DISABLE interrupts
	ret

;
op_79:	;ADC abs,Y
;
	add	bp,4	;4 Cycles
	abs_y		;Calculate address
	jmp	adc_continue	;Continue as standard ADC A,(nnnn)

;
op_7D:	;ADC abs,X
;
	add	bp,4	;4 Cycles
	abs_x		;Calculate address
	jmp	adc_continue	;Continue as standard ADC A,(nnnn)

;
op_7E:	;ROR abs,X
;
	add	bp,7	;7 Cycles
	abs_x		;Calculate address
	jmp	ror_continue	;Continue as standard ROR (nnnn)

;
op_81:	;STA (ind,X)
;
	add	bp,6	;6 Cycles
	ind_x		;Calc address
	mov	[di],dl	;STA
	ret		;Done

;
op_84:	;STY (zpage)
;
	inc	bp
	inc	bp	;2 Cycles
	zpage		;Calc address
	mov	[di],cl	;STY
	ret		;Done

;
op_85:	;STA (zpage)
;
	inc	bp
	inc	bp	;2 Cycles
	zpage		;Calc address
	mov	[di],dl	;STA
	ret		;Done

;
op_86:	;STX (zpage)
;
	inc	bp
	inc	bp	;2 Cycles
	zpage		;Calc address
	mov	[di],ch	;STX
	ret		;Done

;
op_88:	;DEY
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	dec	cl	;Do the DEY
	jnz	op_88nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_88nz:	mov	al,cl
	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_8A:	;TXA
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	mov	dl,ch	;Accum = X
	mov	al,dl	;Get 6502 Accum
	test	al,al	;Update flags
	jnz	op_8Anz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_8Anz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_8C:	;STY (absolute)
;
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	mov	[di],cl	;STY
	ret		;Done

;
op_8D:	;STA (absolute)
;
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	mov	[di],dl	;STA

	cmp	di,bank
	jnz	no_swap
	mov	al,dl	;Get contents of bankswitch
	and	al,bankbit	;Mask bank bit
	cmp	al,cs:[last_swap]	;Same as last time
	jz	no_swap	;Yes, so no swap
	push	cx
	push	si
	mov	cs:[last_swap],al	;Save this bankswap
	mov	cx,256	;256 bytes
	mov	si,200h	;Player1 vars
	mov	di,300h	;Player2 vars
do_swap:	mov	al,[si]
	mov	ah,[di]	
	xchg	ah,al	;Swap 'em
	mov	[di],ah
	mov	[si],al
	inc	di
	inc	si
	loop	do_swap	;Do remainder of bytes
	mov	al,10h	          ;Beat toggle
	mov byte ptr cs:[beat_toggle+1],al  ;Store it
	pop	si
	pop	cx
	
no_swap:	ret		;Done

;
op_8E:	;STX (absolute)
;
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	mov	[di],ch	;STX
	ret		;Done

;
op_90:	;BCC disp
;
	inc	bp
	inc	bp	;2 Cycles
	test	dh,1	;Zero if carry flag clear
	branch_clear	;Return through branch handler

;
op_91:	;STA (ind),Y
;
	add	bp,6	;6 Cycles
	ind_y		;Calc address
	mov	[di],dl	;STA
	ret		;Done

;
op_94:	;STY zpage,X
;
	add	bp,4	;4 Cycles
	zpage_x		;Calc address
	mov	[di],cl	;STY
	ret		;Done

;
op_95:	;STA zpage,X
;
	add	bp,4	;4 Cycles
	zpage_x		;Calc address
	mov	[di],dl	;STA
	ret		;Done

;
op_96:	;STX zpage,Y
;
	add	bp,4	;4 Cycles
	zpage_y		;Calc address
	mov	[di],ch	;STX
	ret		;Done

;
op_98:	;TYA
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	mov	dl,cl	;Accum = Y
	mov	al,dl	;Get 6502 Accum
	test	al,al	;Update flags
	jnz	op_98nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_98nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_99:	;STA abs,Y
;
	add	bp,5	;5 Cycles
	abs_y		;Calc address
	mov	[di],dl	;STA
	ret		;Done

;
op_9A:	;TXS
;
	inc	bp
	inc	bp	;2 Cycles
	mov	bl,ch	;SP = X
	ret		;Done

;
op_9D:	;STA abs,X
;
	add	bp,5	;5 Cycles
	abs_x		;Calc address
	mov	[di],dl	;STA
	ret		;Done

;
op_A0:	;LDY immed
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	mov	al,[si]	;Get immediate value
	inc	si
	mov	cl,al	;Set 6502 Y
	test	al,al	;Update flags
	jnz	op_A0nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_A0nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_A1:	;LDA (ind,X)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,6	;6 Cycles
	ind_x		;Calc address
	mov	al,[di]	;Load ACCUM
	mov	dl,al	;Set 6502 Accum
	test	al,al	;Update flags
	jnz	op_A1nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_A1nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_A2:	;LDX immed
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	mov	al,[si]	;Get immediate value
	mov	ch,al	;Set 6502 X
	inc	si
	test	al,al	;Update flags
	jnz	op_A2nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_A2nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_A4:	;LDY (zpage)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,3	;3 Cycles
	zpage		;Calc address
	mov	al,[di]	;Get Y
	mov	cl,al	;Set 6502 Y
	test	al,al	;Update flags
	jnz	op_A4nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_A4nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_A5:	;LDA (zpage)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,3	;3 Cycles
	zpage		;Calc address
	mov	al,[di]	;Get A
	mov	dl,al	;Set 6502 Accum
	test	al,al	;Update flags
	jnz	op_A5nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_A5nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_A6:	;LDX (zpage)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,3	;3 Cycles
	zpage		;Calc address
	mov	al,[di]	;Get X
	mov	ch,al	;Set 6502 X
	test	al,al	;Update flags
	jnz	op_A6nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_A6nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_A8:	;TAY
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	mov	cl,dl	;Accum to Y
	mov	al,cl	;Get Y
	test	al,al	;Update flags
	jnz	op_A8nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_A8nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_A9:	;LDA immed
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	mov	al,[si]	;Get immediate value
	mov	dl,al	;Set 6502 Accum
	inc	si
	test	al,al	;Update flags
	jnz	op_A9nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_A9nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_AA:	;TAX
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	mov	ch,dl	;X = Accum
	mov	al,ch	;Get X
	test	al,al	;Update flags
	jnz	op_AAnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_AAnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_AC:	;LDY (absolute)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	mov	al,[di]	;Get Y
	mov	cl,al	;Set 6502 Y
	test	al,al	;Update flags
	jnz	op_ACnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_ACnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_AD:	;LDA (absolute)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	mov	al,[di]	;Get A
	mov	dl,al	;Set 6502 Accum
	test	al,al	;Update flags
	jnz	op_ADnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_ADnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_AE:	;LDX (absolute)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	mov	al,[di]	;Get X
	mov	ch,al	;Set 6502 X
	test	al,al	;Update flags
	jnz	op_AEnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_AEnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_B0:	;BCS disp
;
	inc	bp
	inc	bp	;2 Cycles
	test	dh,1	;Non zero if carry flag set
	branch_set	;Return through branch handler

;
op_B1:	;LDA (ind),Y
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,5	;5 Cycles
	ind_y		;Calc address
	mov	al,[di]	;Get A
	mov	dl,al	;Set 6502 Accum
	test	al,al	;Update flags
	jnz	op_B1nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_B1nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_B4:	;LDY zpage,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	zpage_x		;Calc address
	mov	al,[di]	;Get Y
	mov	cl,al	;Set 6502 Y
	test	al,al	;Update flags
	jnz	op_B4nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_B4nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_B5:	;LDA zpage,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	zpage_x		;Calc address
	mov	al,[di]	;Get A
	mov	dl,al	;Set 6502 Accum
	test	al,al	;Update flags
	jnz	op_B5nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_B5nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_B6:	;LDX zpage,Y
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	zpage_y		;Calc address
	mov	al,[di]	;Get X
	mov	ch,al	;Set 6502 X
	test	al,al	;Update flags
	jnz	op_B6nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_B6nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_B8:	;CLV
;
	inc	bp
	inc	bp	;2 Cycles
	and	dh,10111111b	;Clear overflow flag
	ret

;
op_B9:	;LDA abs,Y
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	abs_y		;Calc address
	mov	al,[di]	;Get A
	mov	dl,al	;Set 6502 Accum
	test	al,al	;Update flags
	jnz	op_B9nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_B9nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_BA:	;TSX
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	mov	ch,bl	;X = SP
	mov	al,ch	;Get X
	test	al,al	;Update flags
	jnz	op_BAnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_BAnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_BC:	;LDY abs,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	abs_x		;Calc address
	mov	al,[di]	;Get Y
	mov	cl,al	;Set 6502 Y
	test	al,al	;Update flags
	jnz	op_BCnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_BCnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_BD:	;LDA abs,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	abs_x		;Calc address
	mov	al,[di]	;Get A
	mov	dl,al	;Set 6502 Accum
	test	al,al	;Update flags
	jnz	op_BDnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_BDnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_BE:	;LDX abs,Y
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,4	;4 Cycles
	abs_y		;Calc address
	mov	al,[di]	;Get X
	mov	ch,al	;Set 6502 X
	test	al,al	;Update flags
	jnz	op_BEnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_BEnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_C0:	;CPY immed
;
	inc	bp
	inc	bp	;2 Cycles
	mov	al,[si]	;Get immediate value
	inc	si
	cmp	cl,al	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_C1:	;CMP (ind,X)
;
	add	bp,6	;6 Cycles
	ind_x		;Calculate address
	cmp	dl,[di]	;Do CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_C4:	;CPY (zpage)
;
	add	bp,3	;3 Cycles
	zpage		;Calculate address
	cmp	cl,[di]	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_C5:	;CMP (zpage)
;
	add	bp,3	;3 Cycles
	zpage		;Calculate address
	cmp	dl,[di]	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_C6:	;DEC (zpage)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,5	;5 Cycles
	zpage		;Calculate address
	mov	al,[di]	;Get BYTE
	dec	al	;Do the DEC
	mov	[di],al	;Store
	jnz	op_C6nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_C6nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_C8:	;INY
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	inc	cl	;INY
	mov	al,cl	;Get Y
	jnz	op_C8nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_C8nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_C9:	;CMP immed
;
	inc	bp
	inc	bp	;2 Cycles
	mov	al,[si]	;Get immediate value
	inc	si
	cmp	dl,al	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_CA:	;DEX
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	dec	ch	;Do the DEX
	mov	al,ch	;Get X
	jnz	op_CAnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_CAnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_CC:	;CPY (absolute)
;
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	cmp	cl,[di]	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_CD:	;CMP (absolute)
;
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	cmp	dl,[di]	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_CE:	;DEC (absolute)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,6	;6 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	mov	al,[di]	;Get BYTE
	dec	al	;Do the DEC
	mov	[di],al	;Store byte
	jnz	op_CEnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_CEnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_D0:	;BNE disp
;
	inc	bp
	inc	bp	;2 Cycles
	test	dh,2	;Zero if ZERO flag clear
	branch_clear	;Return through branch handler

;
op_D1:	;CMP (ind),Y
;
	add	bp,5	;5 Cycles
	ind_y		;Calculate address
	cmp	dl,[di]	;Do CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_D5:	;CMP zpage,X
;
	add	bp,4	;4 Cycles
	zpage_x		;Calculate address
	cmp	dl,[di]	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_D6:	;DEC zpage,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,6	;6 Cycles
	zpage_x		;Calculate address
	mov	al,[di]	;Get BYTE
	dec	al	;Do the DEC
	mov	[di],al	;Store byte
	jnz	op_D6nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_D6nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_D8:	;CLD
;
	inc	bp
	inc	bp	;2 Cycles
	and	dh,11110111b	;Clear DECIMAL flag
	ret

;
op_D9:	;CMP abs,Y
;
	add	bp,4	;4 Cycles
	abs_y		;Calculate address
	cmp	dl,[di]	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_DD:	;CMP abs,X
;
	add	bp,4	;4 Cycles
	abs_x		;Calculate address
	cmp	dl,[di]	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_DE:	;DEC abs,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,7	;7 Cycles
	abs_x		;Calculate address
	mov	al,[di]	;Get BYTE
	dec	al	;Do the DEC
	mov	[di],al	;Store byte
	jnz	op_DEnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_DEnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_E0:	;CPX immed
;
	inc	bp
	inc	bp	;2 Cycles
	mov	al,[si]	;Get immediate value
	inc	si
	cmp	ch,al	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_E1:	;SBC (ind,X)
;
	add	bp,6	;6 Cycles
	ind_x		;Calc address
	jmp	sbc_continue	;Continue as SBC (absolute)

;
op_E4:	;CPX (zpage)
;
	add	bp,3	;3 Cycles
	zpage		;Calculate address
	cmp	ch,[di]	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_E5:	;SBC (zpage)
;
	add	bp,3	;3 Cycles
	zpage		;Calculate address
	jmp	sbc_continue	;Continue as SBC (absolute)

;
op_E6:	;INC (zpage)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,5	;5 Cycles
	zpage		;Get address
	mov	al,[di]	;Get BYTE
	inc	al	;Do the INC
	mov	[di],al	;Store byte
	jnz	op_E6nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_E6nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_E8:	;INX
;
	and	dh,01111101b	;Clear sign and zero
	inc	bp
	inc	bp	;2 Cycles
	inc	ch	;INX
	mov	al,ch	;Get X
	jnz	op_E8nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_E8nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_E9:	;SBC immed
;
	and	dh,10111111b	;Clear overflow
	inc	bp
	inc	bp	;2 Cycles
	mov	ah,dh	;Set AH = 6502 flags
	test	ah,8	;Doing BCD?
	jnz	op_E9bcd	;Yes
	mov	al,[si]	;Get immediate value
	inc	si
	shr	ah,1	;Move 6502 Carry to 8086's
	cmc		;Compliment it (odd logic!!!)
	sbb	dl,al	;Do the SBC
	cmc		;Compliment it (odd logic!!!)
	flag_s_z_c_o      ;Done, return setting 6502 flags
op_E9bcd:	mov	al,[si]	;Get immediate value
	inc	si
	shr	ah,1	;Move 6502 Carry to 8086's
	cmc		;Compliment it (odd logic!!!)
	sbb	dl,al	;Do the SBC
	mov	al,dl
	das		;Decimal adjust
	cmc		;Compliment it (odd logic!!!)
	mov	dl,al	;8086 accum to 6502's
	flag_s_z_c_o	;Done, return setting 6502 flags

;
op_EC:	;CPX (absolute)
;
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	cmp	ch,[di]	;Do the CMP
	cmc		;For 6502's weird SUB logic!
	flag_sign_zero_c	;Done, return updating S,Z,C flags

;
op_ED:	;SBC (absolute)
;
	and	dh,10111111b	;Clear overflow
	add	bp,4	;4 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
sbc_continue:	mov	ah,dh	;Set AH = 6502 flags
	test	ah,8	;Doing BCD?
	jnz	op_EDbcd	;Yes
	shr	ah,1	;Move 6502 Carry to 8086's
	cmc		;Compliment it (odd logic!!!)
	sbb	dl,[di]	;Do the SBC
	cmc		;Compliment it (odd logic!!!)
	flag_s_z_c_o	;Done, return setting 6502 flags
op_EDbcd:	shr	ah,1	;Move 6502 Carry to 8086's
	cmc		;Compliment it (odd logic!!!)
	sbb	dl,[di]	;Do the SBC
	mov	al,dl	;8086 accum to 6502's
	das		;Decimal adjust
	cmc		;Compliment it (odd logic!!!)
	mov	dl,al	;Back to 6502 accum
	flag_s_z_c_o	;Done, return setting 6502 flags

;
op_EE:	;INC (absolute)
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,6	;6 Cycles
	mov	di,[si]	;Get address
	inc	si
	inc	si
	mov	al,[di]	;Get BYTE
	inc	al	;Do the INC
	mov	[di],al	;Store byte
	jnz	op_EEnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_EEnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_F0:	;BEQ disp
;
	inc	bp
	inc	bp	;2 Cycles
	test	dh,2	;Sets NZ if Zero flag SET
	branch_set	;Return through branch handler

;
op_F1:	;SBC (ind),Y
;
	add	bp,5	;5 Cycles
	ind_y		;Calc address
	jmp	sbc_continue	;Continue as SBC (absolute)

;
op_F5:	;SBC zpage,X
;
	add	bp,4	;4 Cycles
	zpage_x		;Calc address
	jmp	sbc_continue	;Continue as SBC (absolute)

;
op_F6:	;INC zpage,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,6	;6 Cycles
	zpage_x		;Calc address
	mov	al,[di]	;Get BYTE
	inc	al	;Do the INC
	mov	[di],al	;Store byte
	jnz	op_F6nz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_F6nz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
op_F8:	;SED
;
	or	dh,00001000b	;SET decimal mode
;
op_EA:	;NOP
;
	inc	bp
	inc	bp	;2 Cycles
crash:	ret

;
op_F9:	;SBC abs,Y
;
	add	bp,4	;4 Cycles
	abs_y		;Calc address
	jmp	sbc_continue	;Continue as SBC (absolute)

;
op_FD:	;SBC abs,X
;
	add	bp,4	;4 Cycles
	abs_x		;Calc address
	jmp	sbc_continue	;Continue as SBC (absolute)

;
op_FE:	;INC abs,X
;
	and	dh,01111101b	;Clear sign and zero
	add	bp,7	;7 Cycles
	abs_x		;Calc address
	mov	al,[di]	;Get BYTE
	inc	al	;Do the INC
	mov	[di],al	;Store byte
	jnz	op_FEnz	;No ZERO set
	or	dh,00000010b	;Set ZERO flag
op_FEnz:	and	al,10000000b	;Mask SIGN
	or	dh,al	;SET/RES sign
	ret		;Done

;
	include	data.asm	;Include data
;

	END
