;
;A few macros!!
;
s6502	MACRO
	push	bx	;SP
	push	cx	;X and Y
	push	dx	;Flags and Accumulator
	push	si	;PC
	push	bp	;Cycle counter
	ENDM

r6502	MACRO
	pop	bp	;Cycle counter
	pop	si	;PC
	pop	dx	;Flags and Accumulator
	pop	cx	;X and Y
	pop	bx	;SP
	ENDM

;
;6502 Address mode macros, read next part of instrction and build addr in DI
;
ind_x	MACRO
	mov	al,[si]	;Zero page address
	inc	si
	xor	ah,ah	;Zero page!!
	add	al,ch	;Add in X
	mov	di,ax	;First pointer!
	mov	al,[di]
	inc	di
	and	di,0FFh
	mov	ah,[di]
	mov	di,ax	;Get actual pointer!
	ENDM

ind_y	MACRO
	mov	al,[si]	;Get zero page address
	inc	si
	xor	ah,ah	;Zero page!!
	mov	di,ax	;Pointer
	mov	al,[di]	;Base address
	inc	di
	and	di,0FFh
	mov	ah,[di]
	add	al,cl	;Add Y offset
	adc	ah,0	;Add in carry
	mov	di,ax	;Pointer
	ENDM

zpage	MACRO
	mov	al,[si]	;Get zero page low byte
	inc	si
	xor	ah,ah	;Zero page!!
	mov	di,ax	;Set pointer
	ENDM

abs_y	MACRO
	mov	ax,[si]	;Get base address
	inc	si
	inc	si
	add	al,cl	;Add Y
	adc	ah,0	;Add in carry
	mov	di,ax	;Set pointer
	ENDM

abs_x	MACRO
	mov	ax,[si]	;Get base address
	inc	si
	inc	si
	add	al,ch	;Add X
	adc	ah,0	;Add in carry
	mov	di,ax	;Set pointer
	ENDM

zpage_x	MACRO
	mov	al,[si]	;Get zero page
	inc	si
	xor	ah,ah	;Zero page!!
	add	al,ch	;Add X
	mov	di,ax	;Set pointer
	ENDM

zpage_y	MACRO
	mov	al,[si]	;Get zero page
	inc	si
	xor	ah,ah	;Zero page!!
	add	al,cl	;Add Y
	mov	di,ax	;Set pointer
	ENDM

;
;Relative conditional branch handler macros
;
branch_set	MACRO
	LOCAL	do_branch_set

	lodsb		;Get displacement
	jnz	do_branch_set	;Do the branch
	ret		;Done, no flag changes
do_branch_set:	inc	bp
	inc	bp	;Extra 2 Cycles for branch
	cbw		;Sign extend displacement
	add	si,ax	;Adjust PC by +129 to -126
	ret		;Done, no flag changes
	ENDM

branch_clear	MACRO
	LOCAL	do_branch_clear

	lodsb		;Get displacement
	jz	do_branch_clear	;Do the branch
	ret		;Done, no flag changes
do_branch_clear:	inc	bp
	inc	bp	;Extra 2 Cycles for branch
	cbw		;Sign extend displacement
	add	si,ax	;Adjust PC by +129 to -126
	ret		;Done, no flag changes
	ENDM

;
;Flag routine macros, used to swap 8086 and 6502 flags with each other
;
flag_sign_zero_c	MACRO
	lahf		   ;Lo 8086 flags to AH
	and	dh,01111100b	   ;Mask out 6502 S,Z,C flags
	mov	al,ah	   ;8086 flags to AL
	and	ax,1000000101000000b ;Mask out unwanted flags
	rol	al,3	   ;Roll 8086 Z flag to bit 1 AL
	or	al,ah	   ;Build 6502 mask
	or	dh,al	   ;SET/RES 6502 S, Z flags
	ret		   ;Done
	ENDM

flag_s_z_c_o	MACRO
	LOCAL	no_overflow

	lahf		;Lo 8086 flags to AH
	jno	no_overflow	;Reset V flag
	or	dh,01000000b	;SET V flag
no_overflow:	and	dh,01111100b	   ;Mask out 6502 S,Z,C flags
	mov	al,ah	   ;8086 flags to AL
	and	ax,1000000101000000b ;Mask out unwanted flags
	rol	al,3	   ;Roll 8086 Z flag to bit 1 AL
	or	al,ah	   ;Build 6502 mask
	or	dh,al	   ;SET/RES 6502 S, Z flags
	ret		   ;Done
	ENDM

;
;General equates used by the drawing functions
;
crtc_addr	equ	3D4h   ;base port of the CRT controller (colour)
sequ_addr	equ	3C4h   ;base port of the CRT sequencer
colour_number	equ	3C8h   ;VGA pallette number
packed_pix	equ	3CEh   ;base port to manipulate packed pixel mode

;
;For the clipper
;
min_x	equ	0	;Left   ATARI VECTOR = 0
max_x	equ	1023	;Right  ATARI VECTOR = 1023 ????
min_y	equ	0	;Top    ATARI VECTOR = 0
max_y	equ	775	;Bottom ATART VECTOR = 775 ????

drawlist_add	equ	16382	;Toggler
erase_list	equ	16384	;Toggler

;
;Asteroids Machine equates
;
bank	equ	3200h	;200h-300h Bank switch port
bankbit	equ	4	;Bit number of bank switcher
nmicycle	equ	6250	;NMI every 6250 cycles of the 6502
vecgo	equ	3000h	;Vector GO port

;
;Sound Sample length equates, all + 1 byte
;
beat1_length	equ	822
beat2_length	equ	859
lexplode_length	equ	9518
mexplode_length	equ	10851
sexplode_length	equ	9558
pfire_length	equ	2718
sfire_length	equ	2512
smsaucer_length	equ	1370
lgsaucer_length	equ	1858
thrust_length	equ	962
extra_length	equ	1430
